import sys
import asyncio
import contextlib

@contextlib.contextmanager
def sync_contextmanager(async_contextmanager):
  loop = asyncio.get_event_loop()
  ctx = loop.run_until_complete(async_contextmanager.__aenter__())
  try:
    yield ctx
  except:
    loop.run_until_complete(async_contextmanager.__aexit__(*sys.exc_info()))
  else:
    loop.run_until_complete(async_contextmanager.__aexit__(None, None, None))



import logging
logger = logging.getLogger(__name__)

async def try_n_times(n, coro, *args, **kwargs):
  import asyncio
  backoff = 1
  while n > 0:
    try:
      return await coro(*args, **kwargs)
    except Exception as err:
      n -= 1
      if n == 0:
        raise err
      logger.warn(f"{coro} failed, trying again in {backoff}s...")
      await asyncio.sleep(backoff)
      backoff *= 2



def split_protocol_opts(url, default_protocol='file'):
  ''' Like `fsspec.core.split_protocol`
  but remove fragments potentially parsing querystrings in the fragment as filesystem opts

  url of the form: proto://netloc/path?qs=anything#ignored?protocol.options=here
  '''
  from appyter.ext.urllib import parse_file_uri
  from fsspec.core import url_to_fs, split_protocol
  uri_parsed = parse_file_uri(url)
  opts = uri_parsed.fragment_qs or {}
  uri_parsed.fragment = None
  uri_parsed.fragment_query = None
  protocol, path = split_protocol(str(uri_parsed))
  protocol = protocol or default_protocol
  return protocol, path, opts



''' Bare profile with no styling.
```eval_rst
The default profile contains the :mod:`appyter.profiles.default.fields`, :mod:`appyter.profiles.default.filters`, and :mod:`appyter.profiles.default.templates` usable by all other profiles.
```
'''


import logging

logger = logging.getLogger(__name__)

import fsspec
from appyter.ext.fsspec.core import url_to_chroot_fs

class AliasFileSystemBase(fsspec.AbstractFileSystem): pass

def AliasFileSystemFactory(_proto, _fs_url, **_kwargs):
  logger.debug(f"creating AliasFileSystem {_proto}://* => {_fs_url}*")
  class AliasFileSystem(AliasFileSystemBase):
    ''' alias: seemless passthrough to a more elaborate protocol
    '''
    root_marker = ''
    protocol = _proto
    fs_url = _fs_url
    kwargs = _kwargs

    def __init__(self, **kwargs):
      super().__init__(**kwargs)
      self.fs = url_to_chroot_fs(_fs_url, **_kwargs)
      self.root_marker = self.fs.root_marker

    def __enter__(self):
      if getattr(self.fs, '__enter__', None) is not None:
        self.fs.__enter__()
      return self
    
    def __exit__(self, type, value, traceback):
      if getattr(self.fs, '__exit__', None) is None:
        self.fs.__exit__(type, value, traceback)

    def mkdir(self, path, **kwargs):
      return self.fs.mkdir(path, **kwargs)

    def makedirs(self, path, exist_ok=False):
      return self.fs.makedirs(path, exist_ok=exist_ok)

    def rmdir(self, path):
      return self.fs.rmdir(path)

    def rm(self, path, recursive=False, maxdepth=None):
      return self.fs.rm(path, recursive=recursive, maxdepth=maxdepth)

    def copy(self, path1, path2, recursive=False, on_error=None, **kwargs):
      return self.fs.copy(path1, path2, recursive=recursive, on_error=on_error, **kwargs)

    def mv(self, path1, path2, recursive=False, maxdepth=None, **kwargs):
      return self.fs.mv(path1, path2, recursive=recursive, maxdepth=maxdepth, **kwargs)

    def info(self, path, **kwargs):
      return self.fs.info(path, **kwargs)

    def ls(self, path, detail=True, **kwargs):
      return self.fs.ls(path, detail=detail, **kwargs)

    def _open(self, path, mode="rb", block_size=None, cache_options=None, **kwargs):
      return self.fs._open(path, mode=mode, block_size=block_size, cache_options=cache_options, **kwargs)

  return AliasFileSystem

def dump_aliases():
  return {
    k: [[cls.protocol, cls.fs_url], cls.kwargs]
    for k, cls in fsspec.registry.target.items()
    if issubclass(cls, AliasFileSystemBase)
  }

def register_aliases(alias_dump):
  for k, (args, kwargs) in alias_dump.items():
    fsspec.register_implementation(k, AliasFileSystemFactory(*args, **kwargs))



import fsspec
from pathlib import Path
from appyter.ext.urllib import join_url
from appyter.ext.tempfile import tempdir

def serve(app_path, **kwargs):
  import os
  import sys
  import logging
  logger = logging.getLogger(__name__)
  from subprocess import Popen
  from appyter.context import get_env, get_jinja2_env, find_blueprints, get_appyter_directory
  config = get_env(**kwargs)
  logger.info(kwargs)
  env = get_jinja2_env(config=config)
  exit_code = 1
  with tempdir() as tmp_dir:
    logger.info(f"Working directory {tmp_dir}")
    #
    logger.info(f"Pre-rendering pages...")
    with fsspec.open(join_url(config['CWD'], config['IPYNB']), 'r') as fr:
      from appyter.parse.nb import nb_from_ipynb_io
      nbtemplate = nb_from_ipynb_io(fr)
    with (tmp_dir/'index.html').open('w') as fw:
      from appyter.render.form import render_form_from_nbtemplate
      fw.write(render_form_from_nbtemplate(env, nbtemplate))
    with (tmp_dir/'index.json').open('w') as fw:
      import json
      from appyter.render.nbinspect import render_nbtemplate_json_from_nbtemplate
      json.dump(render_nbtemplate_json_from_nbtemplate(env, nbtemplate), fw)
    with (tmp_dir/'landing.html').open('w') as fw:
      env.get_template('landing.j2').stream(
        _nb=os.path.basename(config['IPYNB']),
      ).dump(fw)
    #
    logger.info(f"Generating production config...")
    with (tmp_dir/'supervisord.conf').open('w') as fw:
      env.get_template('production/supervisord.conf.j2').stream(_tmp_dir=tmp_dir, sys=sys, str=str).dump(fw)
    with (tmp_dir/'nginx.conf').open('w') as fw:
      env.get_template('production/nginx.conf.j2').stream(
        _tmp_dir=tmp_dir, os=os, str=str,
        get_appyter_directory=get_appyter_directory,
        find_blueprints=find_blueprints,
      ).dump(fw)
    logger.info(f"Starting production instance at http://{config['HOST']}:{config['PORT']}{config['PREFIX']} ...")
    with Popen(['supervisord', '-n', '-c', str(tmp_dir/'supervisord.conf')]) as proc:
      try:
        exit_code = proc.wait()
      except KeyboardInterrupt:
        proc.terminate()
        exit_code = proc.wait()
  #
  sys.exit(exit_code)



from appyter.fields import Field
from appyter.ext.dict import dict_collision_free_update

class TabField(Field):
  ''' Representing a tab field which contains inner fields to choose from.
  ```python
  %%appyter hide_code
  {% myfield = TabField(
    name='my_field',
    choices={
      "A": [
        StringField(
          name="my_field_A",
          default="a"
        ),
      ],
      "B": [
        StringField(
          name="my_field_B",
          default="b"
        )
      ]
    }
  ) %}

  :param name: (str) A name that will be used to refer to the object as a variable and in the HTML form.
  :param label: (str) A human readable label for the field for the HTML form
  :param description: (Optional[str]) A long human readable description for the field for the HTML form
  :param \**kwargs: Additional keyword arguments used by other fields
  ```
  '''
  def __init__(self, choices={}, **kwargs):
    super(TabField, self).__init__(choices=choices, **kwargs)
    assert self.choices and type(self.choices) == dict, 'TabField requires choices'
    for section, fields in self.choices.items():
      for field in fields:
        field['args']['parent'] = self.args['name']

  def prepare(self, req):
    data = super().prepare(req)
    for value in self.value:
      dict_collision_free_update(data, **value.prepare(req))
    return data

  def constraint(self):
    if not self.raw_value:
      return not self.args.get('required')
    else:
      return all(v in self.choices for v in self.raw_value)



from appyter.fields import Field

class IntField(Field):
  ''' Representing a field that accepts an integer

  :param name: (str) A name that will be used to refer to the object as a variable and in the HTML form.
  :param label: (str) A human readable label for the field for the HTML form
  :param description: (Optional[str]) A long human readable description for the field for the HTML form
  :param min: (Optional[int]) the minimum valid value that the field can take on
  :param max: (Optional[int]) the maximum valid value that the field can take on
  :param step: (Optional[int]) the interval for which values are incremented or decremented
  :param required: (Optional[bool]) Whether or not this field is required (defaults to false)
  :param default: (float) A default value as an example and for use during prototyping
  :param section: (Optional[str]) The name of a SectionField for which to nest this field under, defaults to a root SectionField
  :param value: (INTERNAL Any) The raw value of the field (from the form for instance)
  :param \**kwargs: Remaining arguments passed down to :class:`appyter.fields.Field`'s constructor.
  '''
  def __init__(self, min=None, max=None, step=None, **kwargs):
    super().__init__(
      min=min,
      max=max,
      step=step,
      **kwargs,
    )

  @property
  def raw_value(self):
    if self.args['value'] is None:
      return None
    else:
      return int(self.args['value'])

  def constraint(self):
    if self.raw_value is None:
      return not self.args.get('required')
    else:
      return (
        self.args['min'] is None or self.raw_value >= self.args['min']
      ) and (
        self.args['max'] is None or self.raw_value <= self.args['max']
      ) and (
        self.args['step'] is None or (self.raw_value - self.args.get('min', 0)) % self.args['step'] == 0
      )



''' Orchestration dispatcher implementations using different backends
'''

import os
from appyter.ext.importlib import importdir
from appyter.orchestration.cli import orchestration

# capture all dispatchers in the directory and add to dispatchers/globals
dispatchers = {}
importdir(os.path.dirname(__file__), __package__, dispatchers)
globals().update(**dispatchers)

@orchestration.command(help='List the dispatchers currently available')
def list_dispatchers():
  dispatcher_name_max_size = max(map(len, dispatchers.keys()))
  for dispatcher, mod in dispatchers.items():
    doc_first_line = mod.__doc__.strip().splitlines()[0].strip()
    print(dispatcher.ljust(dispatcher_name_max_size+1), doc_first_line)




def exception_as_dict(exc):
  if getattr(exc, 'as_dict', None) is None:
    return dict(cls=exc.__class__.__name__, message=str(exc))
  else:
    return dict(cls=exc.__class__.__name__, **exc.as_dict())



''' ```eval_rst
This module contains :class:`appyter.fields.Field`, the base class for all fields
defined in :mod:`appyter.profiles.default.fields`.
``` '''

from flask import Markup
from appyter.ext.itertools import collapse

class PartialField:
  ''' Partial instantiation of a field
  Replaces a decorator so that we can still identify it as
   a callable which will produce a field.
  '''
  def __init__(self, field, **kwargs):
    self._field = field
    self._kwargs = kwargs

  def __call__(self, name=None, value=None, **kwargs):
    kwargs = dict(self._kwargs, **kwargs)
    value = kwargs.pop('context').get(name) if value is None else value
    return self._field(name=name, value=value, **kwargs)

def build_fields(fields, context={}, env=None):
  ''' INTERNAL: Build a dictionary of Field instances
  '''
  return {
    field_name: PartialField(field, context=context, _env=env)
    for field_name, field in fields.items()
  }

class FieldConstraintException(Exception):
  def __init__(self, field, field_name, value, message=None):
    self.field = field
    self.field_name = field_name
    self.value = value
    if message is None: message = "{}[{}]: {} does not satisfy constraints".format(field, field_name, repr(value))
    self.message = message
    super().__init__(message)
  
  def as_dict(self):
    return dict(
      field=self.field,
      field_name=self.field_name,
      value=self.value,
      message=self.message,
    )

class Field(dict):
  ''' Base field for which all fields derive
  ```eval_rst
  Base class for all Field objects representing a value that will later be provided via a front-end form.
  See :mod:`appyter.profiles.default.fields` for the actual fields.
  ``` '''
  def __init__(self,
      name=None,
      label=None,
      description=None,
      choices=[],
      required=False,
      default=None,
      value=None,
      section=None,
      _env=None,
      **kwargs):
    '''
    :param name: (str) A name that will be used to refer to the object as a variable and in the HTML form.
    :param label: (str) A human readable label for the field for the HTML form
    :param description: (Optional[str]) A long human readable description for the field for the HTML form
    :param choices: (Optional[Union[List[str], Set[str], Dict[str, str]]]) A set of choices that are available for this field or lookup table mapping from choice label to resulting value
    :param required: (Optional[bool]) Whether or not this field is required (defaults to false)
    :param default: (Any) A default value as an example and for use during prototyping
    :param section: (Optional[str]) The name of a SectionField for which to nest this field under, defaults to a root SectionField
    :param value: (INTERNAL Any) The raw value of the field (from the form for instance)
    :param \**kwargs: Additional keyword arguments used by other fields
    '''
    super().__init__(
      field=self.field,
      args=dict(
        name=name,
        label=label,
        description=description,
        choices=choices,
        required=required,
        default=default,
        value=value if value is not None else default,
        section=section,
        **kwargs,
      )
    )
    assert name is not None, "Name should be defined and unique"
    self._env = _env
  
  @property
  def args(self):
    ''' Get the raw args, the values used to initialize this field
    '''
    return self['args']

  def prepare(self, req):
    ''' Given a flask request, capture relevant variables for this field
    '''
    data = {}
    if type(req) == dict:
      value = req.get(self.args['name']) if self.args['name'] in req else self.args.get('default')
    elif req.json:
      value = req.json.get(self.args['name']) if self.args['name'] in req.json else self.args.get('default')
    elif req.form:
      value = collapse(req.form.getlist(self.args['name'])) if self.args['name'] in req.form else self.args.get('default')
    else:
      raise NotImplementedError
    #
    self.args['value'] = value
    return { self.args['name']: value }

  def constraint(self):
    ''' Return true if the received args.value satisfies constraints.
    Should be overridden by subclasses.
    '''
    return (self.raw_value is None and not self.args.get('required')) or (self.raw_value in self.choices)

  def render(self, **kwargs):
    ''' Return a rendered version of the field (form)

    :param \**kwargs: The instance values of the form e.g. `Field.render(**field.args)`
    '''
    return Markup(
      self._env.get_template(
        self.template
      ).render(dict(**kwargs, this=self))
    )

  @property
  def field(self):
    ''' Field name
    '''
    return self.__class__.__name__

  @property
  def template(self):
    ''' Template to use for rendering field
    '''
    return '/'.join(['fields', self.field + '.j2'])

  @property
  def js_url(self):
    ''' Template to use for rendering field
    '''
    from appyter.profiles.default.filters.url_for import url_for
    return url_for('static', filename='js/fields/' + self.field + '.js')

  @property
  def choices(self):
    ''' Potential values to choose from
    '''
    return self.args['choices']

  @property
  def raw_value(self):
    ''' (UNSAFE) Raw value of the field
    '''
    return self.args['value']

  @property
  def value(self):
    ''' (SEMI-SAFE) Effective raw value of the field when parsed and constraints are asserted.
    When instantiating code, you should use safe_value.
    '''
    choices = self.choices
    if self.raw_value is None:
      if not self.args.get('required'):
        return None
      else:
        raise FieldConstraintException(
          field=self.field,
          field_name=self.args['name'],
          value=self.raw_value,
          message='{}[{}] is required'.format(self.field, self.args['name']),
        )
    elif type(choices) == dict:
      if self.raw_value in choices:
        return choices[self.raw_value]
      else:
        raise FieldConstraintException(
          field=self.field,
          field_name=self.args['name'],
          value=self.raw_value,
        )
    elif not self.constraint():
      raise FieldConstraintException(
        field=self.field,
        field_name=self.args['name'],
        value=self.raw_value,
      )
    else:
      return self.raw_value

  @property
  def safe_value(self):
    ''' (SAFE) Effective safe value for use in code, we use `repr` to escape values as necessary
    '''
    return repr(self.value)

  @property
  def render_value(self):
    ''' (SAFE) Effective safe value ready to be displayed, specifically for Markdown output.
    '''
    return Markup(self.value)

  def __str__(self):
    ''' (SAFE) The default str(Field) is just safe_value
    '''
    return self.safe_value



import os
import sys
import click
from appyter.cli import cli
from appyter.ext.click import click_option_setenv
from appyter.ext.tempfile import tempdir

@cli.command(
  help='A simple alias for jupyter notebook which re-uses appyter environment config',
  context_settings=dict(
    ignore_unknown_options=True,
  ),
)
@click_option_setenv('--data-dir', envvar='APPYTER_DATA_DIR', default=None, help='The directory to store data of executions')
@click_option_setenv('--cwd', envvar='APPYTER_CWD', default=os.getcwd(), help='The directory to treat as the current working directory for templates and execution')
@click_option_setenv('--host', envvar='APPYTER_HOST', default='0.0.0.0', help='The host to bind to')
@click_option_setenv('--port', envvar='APPYTER_PORT', type=int, default=5000, help='The port this server should run on')
@click.argument('args', nargs=-1, type=click.UNPROCESSED)
def serve(data_dir, cwd, host, port, args):
  from subprocess import Popen
  # add cwd (the appyter itself) to the PYTHONPATH
  sys.path.insert(0, os.path.realpath(cwd))
  # create blank data_dir as tmpdir if it doesn't exist
  with tempdir(data_dir) as tmp:
    #
    # prepare default flags
    flags = {
      'NotebookApp.ip': host,
      'NotebookApp.port': port,
      'NotebookApp.token': '',
      'NotebookApp.password': '',
    }
    # look through args and replace any overrides
    extra_args = []
    for arg in args:
      if arg.startswith('--'):
        name, value = arg[2:].split('=', maxsplit=1)
        flags[name] = value
      else:
        extra_args.append(arg)
    #
    # run jupyter notebook interactively and exit with it
    proc = Popen(
      [
        'jupyter', 'notebook',
        *[
          f"--{key}={value}"
          for key, value in flags.items()
        ],
        *extra_args,
      ],
      stdin=sys.stdin, stdout=sys.stdout, stderr=sys.stderr,
      cwd=tmp.resolve(),
      env=dict(
        PYTHONPATH=':'.join(sys.path),
        PATH=os.environ['PATH']
      ),
    )
    exit_code = None
    while exit_code is None:
      try:
        exit_code = proc.wait()
      except KeyboardInterrupt:
        exit_code = 0
        break
      except:
        pass
  #
  sys.exit(exit_code)



import traceback
import fsspec
from flask import Blueprint, request, redirect, abort, url_for, current_app, jsonify, make_response

from appyter.context import get_jinja2_env
from appyter.ext.dict import dict_collision_free_update
from appyter.ext.fsspec.core import url_to_chroot_fs
from appyter.ext.urllib import join_url
from appyter.parse.nb import nb_to_ipynb_io
from appyter.ext.exceptions import exception_as_dict
from appyter.render.flask_app.constants import get_fields, get_ipynb_hash, get_j2_env, get_nbtemplate
from appyter.render.nbconstruct import render_nb_from_nbtemplate
from appyter.ext.flask import route_join_with_or_without_slash
from appyter.ext.hashlib import sha1sum_dict

core = Blueprint('__main__', __name__)

def prepare_data(req):
  data = {}
  for field in get_fields():
    dict_collision_free_update(data, **field.prepare(req))
  return data

def prepare_results(data):
  results_hash = sha1sum_dict(dict(ipynb=get_ipynb_hash(), data=data))
  with url_to_chroot_fs(join_url('storage://output/', results_hash)) as data_fs:
    if not data_fs.exists(current_app.config['IPYNB']):
      # construct notebook
      env = get_jinja2_env(config=current_app.config, context=data, session=results_hash)
      nbtemplate = get_nbtemplate()
      # in case of constraint failures, we'll fail here
      nb = render_nb_from_nbtemplate(env, nbtemplate, fields=get_fields(), data=data)
      # write notebook
      with data_fs.open(current_app.config['IPYNB'], 'w') as fw:
        nb_to_ipynb_io(nb, fw)
  #
  return results_hash

@route_join_with_or_without_slash(core, methods=['POST'])
def post_index():
  mimetype = request.accept_mimetypes.best_match([
    'text/html',
    'application/json',
  ], 'text/html')
  #
  try:
    data = prepare_data(request)
    result_hash = prepare_results(data)
    error = None
  except Exception as e:
    traceback.print_exc()
    error = exception_as_dict(e)
  #
  if mimetype in {'text/html'}:
    if error: abort(406)
    else: return redirect(url_for('__main__.data_files', path=result_hash + '/'), 303)
  elif mimetype in {'application/json'}:
    if error is not None:
      return make_response(jsonify(error=error), 406)
    else:
      return make_response(jsonify(session_id=result_hash), 200)
  else:
    abort(404)

@route_join_with_or_without_slash(core, 'ssr', methods=['POST'])
def post_ssr():
  env = get_j2_env()
  try:
    ctx = request.get_json()
    assert ctx['field'].endswith('Field'), 'Invalid field'
    return env.globals[ctx['field']](**ctx['args']).render()
  except Exception as e:
    return make_response(jsonify(error=exception_as_dict(e)), 406)



import multiprocessing as mp
mp.set_start_method('spawn', True)

import tempfile
import contextlib
from pathlib import Path
from fsspec.core import url_to_fs

import appyter.ext.fsspec
from appyter.ext.fsspec.overlayfs import OverlayFileSystem

def assert_eq(a, b): assert a == b, f"{repr(a)} != {repr(b)}"

@contextlib.contextmanager
def _test_ctx():
  with tempfile.TemporaryDirectory() as tmpdir:
    tmpdir = Path(tmpdir)
    (tmpdir/'a'/'b').mkdir(parents=True)
    with (tmpdir/'a'/'b'/'c').open('w') as fw:
      fw.write('C')
    with (tmpdir/'a'/'d').open('w') as fw:
      fw.write('D')
    with (tmpdir/'e').open('w') as fw:
      fw.write('E')
    yield tmpdir

def test_file_overlayfs():
  with _test_ctx() as lower_tmpdir:
    with tempfile.TemporaryDirectory() as upper_tmpdir:
      upper_tmpdir = Path(upper_tmpdir)
      assert_eq(frozenset(str(p.relative_to(lower_tmpdir)) for p in lower_tmpdir.rglob('*')), frozenset(['a', 'a/b', 'a/b/c', 'a/d', 'e']))
      lower_fs, _ = url_to_fs(f"chroot::{lower_tmpdir}")
      upper_fs, _ = url_to_fs(f"chroot::{upper_tmpdir}")
      with OverlayFileSystem(lower_fs=lower_fs, upper_fs=upper_fs) as fs:
        assert_eq(frozenset(fs.glob('**')), frozenset(['a', 'a/b', 'a/b/c', 'a/d', 'e']))
        fs.copy('e', 'E')
        fs.pipe('e', b'A')
        with fs.open('a/d', 'a') as fw:
          fw.write('!')
        assert_eq(fs.cat('a/d'), b'D!')
        assert_eq(fs.cat('e'), b'A')
        assert_eq(fs.cat('E'), b'E')
      assert_eq(frozenset(str(p.relative_to(upper_tmpdir)) for p in upper_tmpdir.rglob('*')), frozenset(['a', 'a/d', 'e', 'E']))
      assert_eq((upper_tmpdir/'e').open('rb').read(), b'A')
      assert_eq((upper_tmpdir/'E').open('rb').read(), b'E')
      assert_eq((upper_tmpdir/'a/d').open('rb').read(), b'D!')
      assert_eq((lower_tmpdir/'a/d').open('rb').read(), b'D')
      assert_eq((lower_tmpdir/'e').open('rb').read(), b'E')



import re
from appyter.fields import Field, FieldConstraintException
from appyter.ext.re import re_full

class TextListField(Field):
  ''' Representing a field that accepts a set of strings separated by newlines

  Be careful with this field, consider defining a constraint regex.
  ```eval_rst
  Unlike :class:`appyter.profiles.default.fields.TextField`, this class will return
   a list and potentially render differently.
  ```

  :param name: (str) A name that will be used to refer to the object as a variable and in the HTML form.
  :param label: (str) A human readable label for the field for the HTML form
  :param description: (Optional[str]) A long human readable description for the field for the HTML form
  :param choices: (Union[List[str], Set[str], Dict[str, str]]) A set of choices that are available for this field or lookup table mapping from choice label to resulting value
  :param constraint: (Regex[str]) A regular expression for validating the file name.
  :param required: (Optional[bool]) Whether or not this field is required (defaults to false)
  :param default: (List[str]) A default value as an example and for use during prototyping
  :param examples: (Optional[Dict[str, List[str]]]) Named lists to provide as clickable examples
  :param hint: (Optional[str]) A hint to put in the field prior to content.
  :param rows: (Optional[int]) The number of rows (lines) in the textarea
  :param cols: (Optional[int]) The number of cols (horizontal characters) in the textarea
  :param section: (Optional[str]) The name of a SectionField for which to nest this field under, defaults to a root SectionField
  :param value: (INTERNAL Any) The raw value of the field (from the form for instance)
  :param \**kwargs: Remaining arguments passed down to :class:`appyter.fields.Field`'s constructor.
  '''
  def __init__(self, constraint=r'[^\n]*', hint=None, **kwargs):
    super().__init__(
      constraint=constraint,
      hint=hint,
      **kwargs,
    )

  @property
  def raw_value(self):
    if type(self.args['value']) == str:
      return list(
        filter(None,
          map(str.strip,
            self.args['value'].splitlines()
          )
        )
      )
    elif type(self.args['value']) == list:
      return self.args['value']
    elif self.args['value'] is None:
      return []
    else:
      return None

  def constraint(self):
    if not self.raw_value:
      return not self.args.get('required')
    else:
      return all([
        re.match(re_full(self.args['constraint']), val)
        for val in self.raw_value
      ])

  @property
  def value(self):
    if not self.constraint():
      raise FieldConstraintException(
        field=self.field,
        field_name=self.args['name'],
        value=self.raw_value,
      )
    elif type(self.choices) == dict:
      return [self.choices[v] for v in self.raw_value]
    else:
      return self.raw_value



''' Represent the areas that can be handled externally in production
'''
import os
from flask import request, current_app, send_file, send_from_directory, abort, jsonify
from werkzeug.exceptions import NotFound

from appyter.ext.urllib import join_url
from appyter.render.flask_app.constants import get_form, get_ipynb_io, get_nbtemplate_json, get_output_fs, get_static_fs, get_j2_env
from appyter.render.flask_app.core import core
from appyter.ext.flask import route_join_with_or_without_slash
from appyter.context import get_appyter_directory

@route_join_with_or_without_slash(core, methods=['GET'])
def get_index():
  mimetype = request.accept_mimetypes.best_match([
    'text/html',
    'application/json',
    'application/vnd.jupyter',
  ], 'text/html')
  if mimetype in {'text/html'}:
    return get_form()
  elif mimetype in {'application/json'}:
    return jsonify(get_nbtemplate_json())
  elif mimetype in {'application/vnd.jupyter'}:
    return send_file(get_ipynb_io(), attachment_filename=current_app.config['IPYNB'], mimetype=mimetype)
  else:
    abort(404)

@core.route('/favicon.ico', methods=['GET'])
def favicon():
  static = get_static_fs()
  if static.exists('favicon.ico'):
    return send_file(static.open('favicon.ico', 'rb'), attachment_filename='favicon.ico')
  abort(404)

@core.route('/static/<path:filename>', methods=['GET'])
def static(filename):
  static = get_static_fs()
  if static.exists(filename):
    return send_file(static.open(filename, 'rb'), attachment_filename=filename)
  #
  try:
    return send_from_directory(get_appyter_directory(f"profiles/{current_app.config['PROFILE']}/static"), path=filename)
  except NotFound:
    return send_from_directory(get_appyter_directory('profiles/default/static'), path=filename)

@route_join_with_or_without_slash(core, '<path:path>', methods=['GET'])
def data_files(path):
  if path.endswith('/'):
    mimetype = request.accept_mimetypes.best_match([
      'text/html',
      'application/json',
      'application/vnd.jupyter',
    ], 'text/html')
    if mimetype == 'text/html':
      return get_j2_env().get_template('landing.j2').render(
        _nb=current_app.config['IPYNB'],
      )
    else:
      output_fs = get_output_fs()
      path = join_url(path, current_app.config['IPYNB'])
      if output_fs.exists(path):
        return send_file(output_fs.open(path, 'rb'), attachment_filename=current_app.config['IPYNB'])
  else:
    output_fs = get_output_fs()
    if output_fs.exists(path):
      return send_file(output_fs.open(path, 'rb'), attachment_filename=os.path.basename(path))
  abort(404)



def main():
  import multiprocessing as mp
  mp.set_start_method('spawn')
  # load dotenv if present
  from dotenv import load_dotenv
  load_dotenv()
  # load everything permitting injection to cli
  import os
  from appyter.ext.importlib import importdir
  importdir(os.path.join(os.path.dirname(__file__)), __package__, globals())
  # grab cli
  from appyter.cli import cli
  cli()

if __name__ == '__main__':
  main()



async def storage_ctx(app):
  data_dir = app['config']['DATA_DIR']
  import fsspec
  from appyter.ext.fsspec.alias import AliasFileSystemFactory
  fsspec.register_implementation('storage', AliasFileSystemFactory('storage', data_dir))
  yield



import os
import glob
import click
import logging
logger = logging.getLogger(__name__)
from appyter.cli import cli
from appyter.ext.click import click_option_setenv, click_argument_setenv
from appyter.ext.json import try_json_loads
from appyter.ext.importlib import importdir_deep
from appyter.ext.urllib import join_slash

def find_fields_dir_mappings(config=None):
  assert config is not None
  cwd = config['CWD']
  profile = config['PROFILE']
  mappings = {}
  if profile != 'default':
    mappings[os.path.join(os.path.dirname(__file__), 'profiles', profile, 'fields') + os.path.sep] = __package__ + '.profiles.' + profile + '.fields'
  mappings[os.path.join(os.path.dirname(__file__), 'profiles', 'default', 'fields') + os.path.sep] = __package__ + '.profiles.default.fields'
  mappings[os.path.abspath(os.path.join(cwd, 'fields')) + os.path.sep] = 'fields'
  return mappings

def find_fields(config=None):
  assert config is not None
  from appyter.fields import Field
  ctx = {}
  for _dirname_, _package_ in find_fields_dir_mappings(config=config).items():
    if os.path.isdir(_dirname_):
      importdir_deep(
        _dirname_,
        _package_,
        ctx,
        filter_mod=lambda m, k, v: not k.startswith('_') and isinstance(v, type) and issubclass(v, Field)
      )
  return ctx

@cli.command(help='List the available fields')
@click_option_setenv('--cwd', envvar='APPYTER_CWD', default=os.getcwd(), help='The directory to treat as the current working directory for templates and execution')
def list_fields(**kwargs):
  fields = find_fields(get_env(ipynb='app.ipynb', **kwargs))
  field_name_max_size = max(map(len, fields.keys()))
  for field, mod in fields.items():
    doc_lines = [
      *(mod.__doc__.strip().splitlines() if mod.__doc__ else []),
      *(mod.__init__.__doc__.strip().splitlines() if mod.__init__.__doc__ else []),
    ]
    doc_first_line = doc_lines[0].strip() if doc_lines else ''
    print(field.ljust(field_name_max_size+1), doc_first_line)

@cli.command(help='Describe a field using its docstring')
@click_option_setenv('--cwd', envvar='APPYTER_CWD', default=os.getcwd(), help='The directory to treat as the current working directory for templates and execution')
@click_argument_setenv('field', envvar='APPYTER_FIELD', type=str)
def describe_field(field, **kwargs):
  fields = find_fields(get_env(ipynb='app.ipynb', **kwargs))
  assert field in fields, 'Please choose a valid field name, see list-fields for options'
  mod = fields[field]
  doc_lines = [
    *(mod.__doc__.strip().splitlines() if mod.__doc__ else []),
    *(mod.__init__.__doc__.strip().splitlines() if mod.__init__.__doc__ else []),
  ]
  print(*doc_lines, sep='\n')


def find_filters_dir_mappings(config=None):
  assert config is not None
  cwd = config['CWD']
  profile = config['PROFILE']
  mappings = {}
  if profile != 'default':
    mappings[os.path.join(os.path.dirname(__file__), 'profiles', profile, 'filters') + os.path.sep] = __package__ + '.profiles.' + profile + '.filters'
  mappings[os.path.join(os.path.dirname(__file__), 'profiles', 'default', 'filters') + os.path.sep] = __package__ + '.profiles.default.filters'
  mappings[os.path.abspath(os.path.join(cwd, 'filters')) + os.path.sep] = 'filters'
  return mappings

def find_filters(config=None):
  assert config is not None
  ctx = {}
  for _dirname_, _package_ in find_filters_dir_mappings(config=config).items():
    if os.path.isdir(_dirname_):
      importdir_deep(
        _dirname_,
        _package_,
        ctx,
        filter_mod=lambda m, k, v: not k.startswith('_') and isinstance(v, type(lambda: None))
      )
  return ctx

def find_blueprints_dir_mappings(config=None):
  assert config is not None
  cwd = config['CWD']
  profile = config['PROFILE']
  mappings = {}
  if profile != 'default':
    mappings[os.path.join(os.path.dirname(__file__), 'profiles', profile, 'blueprints') + os.path.sep] = __package__ + '.profiles.' + profile + '.blueprints'
  mappings[os.path.join(os.path.dirname(__file__), 'profiles', 'default', 'blueprints') + os.path.sep] = __package__ + '.profiles.default.blueprints'
  mappings[os.path.abspath(os.path.join(cwd, 'blueprints')) + os.path.sep] = 'blueprints'
  return mappings

def filter_blueprints(m, k, v):
  from flask import Blueprint
  if isinstance(v, Blueprint):
    return True
  elif callable(v):
    if m.__name__.startswith('blueprints') and 'blueprints.' + k == m.__name__:
      return True
  return False

def find_blueprints(config=None):
  assert config is not None
  ctx = {}
  for _dirname_, _package_ in find_blueprints_dir_mappings(config=config).items():
    if os.path.isdir(_dirname_):
      importdir_deep(
        _dirname_,
        _package_,
        ctx,
        filter_mod=filter_blueprints,
      )
  return ctx

def find_templates_dir(config=None):
  assert config is not None
  return list(filter(os.path.isdir, [
    os.path.abspath(os.path.join(config['CWD'], 'templates')) + os.path.sep,
    os.path.join(os.path.dirname(__file__), 'profiles', config['PROFILE'], 'templates') + os.path.sep,
    os.path.join(os.path.dirname(__file__), 'profiles', 'default', 'templates') + os.path.sep,
  ]))

def get_extra_files(config=None):
  assert config is not None
  dirs = [
    *find_templates_dir(config=config),
    *find_filters_dir_mappings(config=config).keys(),
    *find_fields_dir_mappings(config=config).keys(),
  ]
  paths = {
    os.path.abspath(f) + (os.path.sep if os.path.isdir(f) else '')
    for d in dirs
    for f in ({
      d,
      # all files recursively in these directories
      *glob.glob(os.path.join(d, '**'), recursive=True),
    } - ({
      # exclude directories or files that begin with _
      *glob.glob(os.path.join(d, '_**', '*'), recursive=True),
      *glob.glob(os.path.join(d, '**', '_*'), recursive=True),
    }))
  }
  paths.add(os.path.abspath(os.path.join(config['CWD'], config['IPYNB'])))
  return list(paths)

def get_appyter_directory(path):
  return os.path.abspath(os.path.join(os.path.dirname(__file__), path))

def get_jinja2_env(context={}, config=None, session=''):
  assert config is not None
  #
  import sys
  from appyter.fields import build_fields
  from jinja2 import Environment, ChoiceLoader, FileSystemLoader
  env = Environment(
    extensions=['jinja2.ext.do'],
    loader=ChoiceLoader([
      FileSystemLoader(d)
      for d in find_templates_dir(config=config)
    ]),
  )
  env.filters.update(**find_filters(config=config))
  env.globals.update(**find_filters(config=config))
  env.globals.update(_config=config, _session=session, _os=os)
  env.globals.update(**build_fields(find_fields(config=config), context=context, env=env))
  return env

_mode = None
def get_env_from_kwargs(mode='default', **kwargs):
  global _mode
  if _mode is None:
    _mode = mode
  else:
    mode = _mode
  #
  import os
  import sys
  import uuid
  PREFIX = try_json_loads(kwargs.get('prefix', os.environ.get('APPYTER_PREFIX', '/')))
  PROFILE = try_json_loads(kwargs.get('profile', os.environ.get('APPYTER_PROFILE', 'default')))
  EXTRAS = try_json_loads(kwargs.get('extras', os.environ.get('APPYTER_EXTRAS', '')))
  HOST = try_json_loads(kwargs.get('host', os.environ.get('APPYTER_HOST', '127.0.0.1')))
  PORT = try_json_loads(kwargs.get('port', os.environ.get('APPYTER_PORT', 5000)))
  PROXY = try_json_loads(kwargs.get('proxy', os.environ.get('APPYTER_PROXY', False)))
  CWD = try_json_loads(os.path.abspath(kwargs.get('cwd', os.environ.get('APPYTER_CWD', str(os.getcwd())))))
  DATA_DIR = try_json_loads(kwargs.get('data_dir', os.environ.get('APPYTER_DATA_DIR', 'data')))
  DISPATCHER = try_json_loads(kwargs.get('dispatcher', os.environ.get('APPYTER_DISPATCHER')))
  DISPATCHER_URL = try_json_loads(kwargs.get('dispatcher_url', os.environ.get('APPYTER_DISPATCHER_URL')))
  DISPATCHER_IMAGE = try_json_loads(kwargs.get('dispatcher_image', os.environ.get('APPYTER_DISPATCHER_IMAGE')))
  SECRET_KEY = try_json_loads(kwargs.get('secret_key', os.environ.get('APPYTER_SECRET_KEY', str(uuid.uuid4()))))
  DEBUG = try_json_loads(kwargs.get('debug', os.environ.get('APPYTER_DEBUG', 'true')))
  STATIC_DIR = try_json_loads(kwargs.get('static_dir', os.environ.get('APPYTER_STATIC_DIR', os.path.join(CWD, 'static'))))
  STATIC_PREFIX = join_slash(PREFIX, 'static')
  IPYNB = try_json_loads(kwargs.get('ipynb', os.environ.get('APPYTER_IPYNB')))
  #
  if mode == 'default' and (IPYNB is None or not os.path.isfile(os.path.join(CWD, IPYNB))):
    logger.error('ipynb was not found')
  #
  if '://' not in DATA_DIR and not os.path.isabs(DATA_DIR):
    DATA_DIR = os.path.join(CWD, DATA_DIR)
  #
  if '://' not in STATIC_DIR and not os.path.isabs(STATIC_DIR):
    STATIC_DIR = os.path.join(CWD, STATIC_DIR)
  #
  if os.path.abspath(CWD) not in sys.path:
    sys.path.insert(0, CWD)
  if CWD not in os.environ['PATH'].split(':'):
    os.environ['PATH'] = CWD + ':' + os.environ['PATH']
  #
  return dict(
    MODE=mode,
    PREFIX=PREFIX,
    PROFILE=PROFILE,
    EXTRAS=EXTRAS,
    HOST=HOST,
    PORT=PORT,
    PROXY=PROXY,
    CWD=CWD,
    DATA_DIR=DATA_DIR,
    DISPATCHER=DISPATCHER,
    DISPATCHER_URL=DISPATCHER_URL,
    DISPATCHER_IMAGE=DISPATCHER_IMAGE,
    SECRET_KEY=SECRET_KEY,
    DEBUG=DEBUG,
    STATIC_DIR=STATIC_DIR,
    IPYNB=IPYNB,
    STATIC_PREFIX=STATIC_PREFIX,
  )

def get_env_from_click():
  ''' Traverse click context and use params for get_env_from_kwargs
  '''
  click_ctx = click.get_current_context()
  # aggregate all params from all parents
  params = {}
  while click_ctx:
    # preserve leaf-values in the case of conflicts
    params = {
      k: params.get(k, click_ctx.params.get(k))
      for k in (click_ctx.params.keys() | params.keys())
    }
    click_ctx = click_ctx.parent
  # use aggregated params to get env
  return get_env_from_kwargs(**params)

def get_env_from_flask():
  ''' If we're running in flask, current_app.config should be available
  '''
  from flask import current_app
  return current_app.config

config = {}

def get_env(**kwargs):
  ''' Try various methods to grab the application config
  (different based on whether we're in a flask thread/somewhere else)
  '''
  #
  try:
    global config
    config = get_env_from_kwargs(**kwargs)
  except AssertionError:
    pass
  #
  try:
    return get_env_from_flask()
  except Exception:
    pass
  #
  try:
    return get_env_from_click()
  except RuntimeError:
    pass
  except AssertionError:
    pass
  #
  return config



def ensure_callable(var):
  if callable(var):
    return var
  else:
    return lambda: var

_memcache = {}
def memcached(func):
  ''' Cache all calls to this function in memory permanently
  '''
  global _memcache
  if func not in _memcache:
    _memcache[func] = {}
  from functools import wraps
  @wraps(func)
  def wrapper(*args, **kwargs):
    _h = hash((args, frozenset(kwargs.items())))
    if _h not in _memcache[func]:
      _memcache[func][_h] = func(*args, **kwargs)
    return _memcache[func][_h]
  return wrapper



''' Launch job natively using python
'''

import os
import sys
import json
import asyncio

async def dispatch(job=None, **kwargs):
  proc = await asyncio.create_subprocess_exec(*[
    sys.executable, '-u',
    '-m', 'appyter', 'orchestration', 'job', json.dumps(job)
  ], env=dict(
    PYTHONPATH=':'.join(sys.path),
    PATH=os.environ['PATH'],
  ), stdout=sys.stdout, stderr=sys.stderr)
  return await proc.wait()



import os
from pathlib import PurePath
from fsspec import filesystem, AbstractFileSystem
from appyter.ext.tempfile import mktemp

import logging
logger = logging.getLogger(__name__)

class WriteCacheFileSystem(AbstractFileSystem):
  protocol = 'writecache'

  def __init__(self, target_protocol=None, target_options=None, fs=None, **kwargs):
    '''
    Parameters
    ----------
    target_protocol: str (optional)
      Target filesystem protocol. Provide either this or ``fs``.
    target_options: dict or None
      Passed to the instantiation of the FS, if fs is None.
    fs: filesystem instance
      The target filesystem to run against. Provide this or ``protocol``.
    '''
    super().__init__(**kwargs)
    if not (fs is None) ^ (target_protocol is None):
      raise ValueError(
          "Please provide one of filesystem instance (fs) or"
          " remote_protocol, not both"
      )
    self.kwargs = target_options or {}
    self.target_protocol = (
      target_protocol
      if isinstance(target_protocol, str)
      else (fs.protocol if isinstance(fs.protocol, str) else fs.protocol[0])
    )
    self.fs = fs if fs is not None else filesystem(target_protocol, **self.kwargs)
    self._local_cache = {}
    self._dir_cache = set()

  def __enter__(self):
    if getattr(self.fs, '__enter__', None) is not None:
      self.fs.__enter__()
    return self
  
  def __exit__(self, type, value, traceback):
    if getattr(self.fs, '__exit__', None) is not None:
      self.fs.__exit__(type, value, traceback)

  def mkdir(self, path, **kwargs):
    path = self.fs.root_marker + path.lstrip('/')
    try:
      self.fs.mkdir(path, **kwargs)
    except:
      raise
    else:
      self._dir_cache.add(path)

  def makedirs(self, path, exist_ok=False):
    path = self.fs.root_marker + path.lstrip('/')
    try:
      self.fs.makedirs(path, exist_ok=exist_ok)
    except:
      raise
    else:
      path_split = path.split('/')
      for i in range(1, len(path_split)+1):
        self._dir_cache.add('/'.join(path_split[:i]))

  def rmdir(self, path):
    path = self.fs.root_marker + path.lstrip('/')
    try:
      self.upper_fs.rmdir(path)
    except:
      raise
    else:
      self._dir_cache.discard(path)

  def rm(self, path, recursive=False, maxdepth=None):
    path = self.fs.root_marker + path.lstrip('/')
    try:
      self.fs.rm(path, recursive=recursive, maxdepth=maxdepth)
    except:
      raise
    else:
      if recursive:
        for d in list(self._dir_cache):
          if d.startswith(path):
            self._dir_cache.remove(d)

  def copy(self, path1, path2, recursive=False, on_error=None, **kwargs):
    path1 = self.fs.root_marker + path1.lstrip('/')
    path2 = self.fs.root_marker + path2.lstrip('/')
    try:
      self.fs.copy(path1, path2, recursive=recursive, on_error=on_error, **kwargs)
    except:
      raise
    else:
      if recursive:
        p1 = PurePath(path1)
        p2 = PurePath(path2)
        for d in list(self._dir_cache):
          p = PurePath(d)
          if p.is_relative_to(p1):
            self._dir_cache.add(str(p2 / p.relative_to(path1)))


  def mv(self, path1, path2, recursive=False, maxdepth=None, **kwargs):
    path1 = self.fs.root_marker + path1.lstrip('/')
    path2 = self.fs.root_marker + path2.lstrip('/')
    try:
      self.fs.mv(path1, path2, recursive=recursive, maxdepth=maxdepth, **kwargs)
    except:
      raise
    else:
      if recursive:
        p1 = PurePath(path1)
        p2 = PurePath(path2)
        for d in list(self._dir_cache):
          p = PurePath(d)
          if p.is_relative_to(p1):
            self._dir_cache.remove(d)
            self._dir_cache.add(str(p2 / p.relative_to(path1)))

  def exists(self, path, **kwargs):
    path = self.fs.root_marker + path.lstrip('/')
    if path in self._local_cache or path in self._dir_cache:
      return True
    return self.fs.exists(path, **kwargs)

  def info(self, path, **kwargs):
    path = self.fs.root_marker + path.lstrip('/')
    if path in self._local_cache:
      return self._local_cache[path]
    elif path in self._dir_cache:
      return {'type': 'directory', 'name': path}
    return self.fs.info(path, **kwargs)

  def ls(self, path, detail=False, **kwargs):
    # add local paths
    path = self.fs.root_marker + path.lstrip('/')
    logger.info(f'ls({path})')
    path_split = ('',) if path == self.fs.root_marker else tuple(path.split('/'))
    results = {}
    for local_file_path, f in self._local_cache.items():
      if tuple(local_file_path.split('/'))[:len(path_split)] == path_split:
        if detail:
          results[local_file_path] = f
        else:
          results[local_file_path] = local_file_path
    for f in self.fs.ls(path, detail=detail, **kwargs):
      if detail:
        results[f['name']] = f
      else:
        results[f] = f
    return list(results.values())

  def _open(self, path, mode='rb', **kwargs):
    path = self.fs.root_marker + path.lstrip('/')
    if 'r' not in mode or '+' in mode:
      if path in self._local_cache:
        details = self._local_cache[path]
      else:
        try:
          details = self.fs.info(path)
          if details['type'] == 'directory':
            raise IsADirectoryError(details['name'])
        except FileNotFoundError:
          details = None
      return LocalTempFile(self, path, mode=mode, details=details)
    return self.fs._open(path, mode=mode, **kwargs)

class LocalTempFile:
  """A temporary local file, which will be uploaded on commit"""

  def __init__(self, fs, path, fn=None, mode="wb", autocommit=True, seek=0, details=None):
    fs._local_cache[path] = details if details else {'type': 'file', 'size': 0}
    fn = fn or mktemp()
    self.mode = mode
    self.fn = fn
    # if the file exists on the remote and we need the contents, get them first
    if details and 'a' in mode or ('r' in mode and '+' in mode):
      fs.fs.get(path, fn)
    self.fh = open(fn, mode)
    if seek:
      self.fh.seek(seek)
    self.path = path
    self.fs = fs
    self.closed = False
    self.autocommit = autocommit

  def __reduce__(self):
    # always open in rb+ to allow continuing writing at a location
    return (
      LocalTempFile,
      (self.fs, self.path, self.fn, "rb+", self.autocommit, self.tell()),
    )

  def __enter__(self):
    return self.fh

  def __exit__(self, exc_type, exc_val, exc_tb):
    self.close()

  def close(self):
    self.fh.close()
    self.closed = True
    if self.autocommit:
      self.commit()

  def discard(self):
    self.fh.close()
    self.cleanup()

  def commit(self):
    self.fs.fs.put(self.fn, self.path)
    self.cleanup()

  def cleanup(self):
    os.remove(self.fn)
    if self.path in self.fs._local_cache:
      del self.fs._local_cache[self.path]

  def __getattr__(self, item):
    return getattr(self.fh, item)



from appyter.ext.socketio import AsyncServer

import logging
logger = logging.getLogger(__name__)

socketio = AsyncServer(async_mode='aiohttp')

@socketio.on('connect')
async def _(sid, environ):
  request = environ['aiohttp.request']
  logger.debug(f"connect: {sid}")
  async with socketio.session(sid) as sess:
    sess['request_url'] = f"{request.scheme}://{request.host}{request.path}"
    sess['config'] = request.app['config']

@socketio.on('disconnect')
async def _(sid):
  logger.debug(f"disconnect: {sid}")



import asyncio
import itertools
import logging
logger = logging.getLogger(__name__)

class PriorityQueuedEmitMixin:
  def __init__(self, *args, **kwargs):
    super().__init__(*args, **kwargs)
    try:
      loop = asyncio.get_event_loop()
    except:
      loop = asyncio.new_event_loop()
      asyncio.set_event_loop(loop)
    #
    self._emit_counter = iter(itertools.count())
    self._emit_queue = asyncio.PriorityQueue()
    self._emit_enabled = asyncio.Event()
    self._emit_dispatcher_task = loop.create_task(self.emit_dispatcher())

  async def emit_dispatcher(self):
    while True:
      _, _, args, kwargs = await self._emit_queue.get()
      await self._emit_enabled.wait()
      try:
        await self.klass.emit(self, *args, **{k:v for k,v in kwargs.items() if v})
      except asyncio.CancelledError:
        raise
      except:
        import traceback
        logger.error(traceback.format_exc())
      finally:
        self._emit_queue.task_done()

  async def _emit(self, evt, data, priority=0, **kwargs):
    await self._emit_queue.put((
      priority, next(self._emit_counter), 
      (evt, data),
      kwargs,
    ))

  async def emit(self, evt, data, priority=0, **kwargs):
    logger.debug('Priority queued emit')
    await self._emit(
      evt, data,
      priority=priority,
      **kwargs,
    )

  async def close(self):
    self._emit_dispatcher_task.cancel()
    try:
      await self._emit_dispatcher_task
    except asyncio.CancelledError:
      pass



import asyncio
import threading

def sync_coro(coro):
  loop = asyncio.get_event_loop()
  res = loop.run_until_complete(coro)
  if threading.current_thread() is not threading.main_thread():
    loop.close()
  return res



import logging
logger = logging.getLogger(__name__)

from fsspec import AbstractFileSystem
from fsspec.core import url_to_fs
from appyter.ext.fsspec.util import split_protocol_opts

class MapperFileSystem(AbstractFileSystem):
  ''' MapperFS is the inverse of a fsspec.mapper -- it lets you use a mapping to
  define a virtual filesystem.
  '''
  root_marker = ''
  protocol = 'mapperfs'

  def __init__(self, pathmap={}, **kwargs):
    '''
    Parameters
    ----------
    pathmap: dict
      A mapping from [path to map onto the target filesystem]: filesystem url to present
    '''
    super().__init__(**kwargs)
    if not pathmap:
      raise ValueError(
          "pathmap is required"
      )
    
    self.pathmap = pathmap
    self.listing = {}
    for mapping in self.pathmap:
      src_split = mapping.split('/')
      for i in range(len(src_split)):
        p = '/'.join(src_split[:i+1])
        if i == 0:
          if '' not in self.listing:
            self.listing[''] = {}
          if p:
            self.listing[''][p] = True
        else:
          self.listing['/'.join(src_split[:i])][p] = True
        if i < len(src_split)-1:
          if p not in self.listing:
            self.listing[p] = {}

  def _pathmap(self, path):
    ''' Return (fs, path) depending on whether we hit a mapped paths or not
    '''
    if path in self.pathmap:
      from fsspec import filesystem
      protocol, path, opts = split_protocol_opts(self.pathmap[path])
      return url_to_fs(f"{protocol}://{path}", **opts)
    else:
      raise FileNotFoundError(path)

  def __enter__(self):
    return self
  
  def __exit__(self, type, value, traceback):
    pass

  def mkdir(self, path, **kwargs):
    raise PermissionError(path)
  
  def makedirs(self, path, exist_ok=False):
    raise PermissionError(path)

  def rmdir(self, path):
    raise PermissionError(path)

  def rm(self, path, recursive=False, maxdepth=None):
    raise PermissionError(path)

  def copy(self, path1, path2, recursive=False, on_error=None, maxdepth=None, **kwargs):
    raise PermissionError(path2)

  def mv(self, path1, path2, recursive=False, maxdepth=None, **kwargs):
    raise PermissionError(path1)

  def exists(self, path, **kwargs):
    return path in self.listing or path in self.pathmap

  def info(self, path, **kwargs):
    if path in self.listing:
      return {
        'name': path,
        'type': 'directory',
      }
    else:
      fs, fs_path = self._pathmap(path)
      info = fs.info(fs_path, **kwargs)
      info = dict(info, name=path)
      return info

  def ls(self, path, detail=False, **kwargs):
    ''' Aggregate results based on pathmap listing & underlying fs
    '''
    logger.debug(f"ls({path})")
    results = []
    if path in self.listing:
      for p in self.listing[path]:
        if detail:
          results.append(self.info(p))
        else:
          results.append(p)
    return results

  def _open(self, path, mode="rb", block_size=None, autocommit=True, cache_options=None, **kwargs):
    fs, fs_path = self._pathmap(path)
    if 'r' not in mode: raise PermissionError(path)
    return fs._open(
      fs_path,
      mode=mode,
      block_size=block_size,
      autocommit=autocommit,
      cache_options=cache_options,
      **kwargs,
    )



import aiohttp
import traceback
import logging
import shutil
from fsspec.core import url_to_fs
from flask import request, jsonify, abort

from appyter.ext.fsspec.core import url_to_chroot_fs
from appyter.render.flask_app.constants import get_input_fs
logger = logging.getLogger(__name__)

from appyter.render.flask_app.core import core
from appyter.render.flask_app.socketio import socketio
from appyter.ext.flask import secure_filepath, secure_url, route_join_with_or_without_slash
from appyter.ext.hashlib import sha1sum_io
from appyter.ext.uuid import generate_uuid

# organize file by content hash
def organize_file_content(data_fs, tmp_fs, tmp_path, filename):
  with tmp_fs.open(tmp_path, 'rb') as fr:
    content_hash = sha1sum_io(fr)
  if not data_fs.exists(content_hash):
    with tmp_fs.open(tmp_path, 'rb') as fr:
      with data_fs.open(content_hash, 'wb') as fw:
        shutil.copyfileobj(fr, fw)
  return f"storage://input/{content_hash}#{filename}"

@route_join_with_or_without_slash(core, 'check', '<path:path>', methods=['GET'])
def check(path):
  qs = request.query_string.decode()
  url = path + (('?'+qs) if qs else '')
  fs, fs_path = url_to_fs(url)
  if fs.exists(fs_path):
    return jsonify(fs.info(fs_path))
  else:
    abort(404)

# upload from client
@socketio.on("siofu_start")
async def siofu_start(sid, data):
  try:
    path = generate_uuid()
    filename = secure_filepath(data.get('name'))
    tmp_fs = url_to_chroot_fs('tmpfs:///')
    async with socketio.session(sid) as sess:
      sess['file_%d' % (data.get('id'))] = dict(
        data,
        path=path,
        name=filename,
        bytesLoaded=0,
        tmp_fs=tmp_fs.__enter__(),
        fh=tmp_fs.open(path, 'wb'),
      )
    await socketio.emit(
      'siofu_ready',
      {
        'id': data.get('id'),
        'name': None,
      },
      room=sid,
    )
  except Exception as e:
    logger.error(traceback.format_exc())
    await socketio.emit('siofu_error', str(e), room=sid)

@socketio.on("siofu_progress")
async def siofu_progress(sid, evt):
  async with socketio.session(sid) as sess:
    sess['file_%d' % (evt['id'])]['fh'].write(evt['content'])
  logger.debug(f"progress: {evt}")
  await socketio.emit("siofu_chunk", dict(
    id=evt['id'],
  ), room=sid)

@socketio.on("siofu_done")
async def siofu_done(sid, evt):
  import asyncio
  async with socketio.session(sid) as sess:
    sess['file_%d' % (evt['id'])]['fh'].close()
    input_fs = get_input_fs()
    tmp_fs = sess['file_%d' % (evt['id'])]['tmp_fs']
    path = sess['file_%d' % (evt['id'])]['path']
    filename = sess['file_%d' % (evt['id'])]['name']
    full_filename = await asyncio.get_event_loop().run_in_executor(None, organize_file_content, input_fs, tmp_fs, path, filename)
    tmp_fs.__exit__(None, None, None)
    del sess['file_%d' % (evt['id'])]
  #
  await socketio.emit('siofu_complete', dict(
    id=evt['id'],
    detail=dict(full_filename=full_filename)
  ), room=sid)

# upload from client with POST
def upload_from_request(req, fname):
  input_fs = get_input_fs()
  fh = req.files.get(fname)
  if not fh:
    return None
  filename = secure_filepath(fh.filename)
  path = generate_uuid()
  with url_to_chroot_fs('tmpfs:///') as tmp_fs:
    with tmp_fs.open(path, 'wb') as fw:
      fh.save(fw)
    return organize_file_content(input_fs, tmp_fs, path, filename)



import asyncio
import logging
logger = logging.getLogger(__name__)

class EventEmitter:
  ''' Javascript-style event emitter
  '''
  def __init__(self):
    self._queues = {}
    self._consumers = {}
    self._listeners = {}
  #
  def _get_queue(self, event):
    if event not in self._queues:
      self._queues[event] = asyncio.Queue()
    return self._queues[event]
  #
  def _consumer_factory(self, event):
    async def _consumer():
      logger.debug(f"Starting consumer for {event}...")
      event_queue = self._get_queue(event)
      while True:
        evt = await event_queue.get()
        logger.debug(f"Calling listeners for {event}...")
        try:
          await asyncio.gather(*[
            listener(**evt)
            for listener in self._listeners.get(event, {}).values()
          ])
        except:
          import traceback
          logger.error(f"Error in listener for {event}: {traceback.format_exc()}")
        finally:
          event_queue.task_done()
    return _consumer
  #
  def _ensure_consumer(self, event):
    if event not in self._consumers:
      self._consumers[event] = asyncio.create_task(self._consumer_factory(event)())
    return self._consumers[event]
  #
  async def emit(self, event, **data):
    logger.debug(f"Emitting {event}: {data}")
    event_queue = self._get_queue(event)
    await event_queue.put(data)
  #
  def on(self, event):
    logger.debug(f"Attaching listener to {event}")
    def wrapper(listener):
      if event not in self._listeners:
        self._listeners[event] = {}
      self._listeners[event][listener] = listener
      self._ensure_consumer(event)
      return listener
    return wrapper
  #
  def off(self, event, listener):
    logger.debug(f"Detaching listener from {event}")
    del self._listeners[event][listener]
  #
  async def wait(self, event):
    available = asyncio.Event()
    result = {}
    @self.on(event)
    async def waiter(**kwargs):
      self.off(event, waiter)
      result[None] = kwargs
      available.set()
    await available.wait()
    return result[None]

  async def flush(self):
    logger.debug(f"Flushing...")
    await asyncio.gather(*[
      event_queue.join()
      for event_queue in self._queues.values()
    ])
  
  async def remove(self, event):
    try:
      await self._queues[event].join()
      self._consumers[event].cancel()
      await self._consumers[event]
    except asyncio.CancelledError:
      pass
    finally:
      del self._consumers[event]

  async def clear(self):
    logger.debug(f"Clearing...")
    await asyncio.gather(*map(self.remove, self._consumers.keys()))



def importdir(_dirname_, _package_, _globals_):
  import os, importlib
  for f in os.listdir(_dirname_):
    if f.startswith('_') or f.startswith('.'):
      continue
    if f.endswith('.py'):
      modname = f[:-len('.py')]
    elif os.path.isdir(os.path.join(_dirname_, f)):
      modname = f
    else:
      continue
    mod = importlib.import_module('.{}'.format(modname), _package_)
    _globals_.update(**{modname: mod})

def importdir_deep(_dirname_, _package_, _globals_, filter_mod=lambda m, k, v: not k.startswith('_')):
  import os, importlib
  for f in os.listdir(_dirname_):
    if f.startswith('_') or f.startswith('.'):
      continue
    if f.endswith('.py'):
      modname = f[:-len('.py')]
    elif os.path.isdir(os.path.join(_dirname_, f)):
      modname = f
    else:
      continue
    mod = importlib.import_module('.{}'.format(modname), _package_)
    _globals_.update(**{
      k: v
      for k, v in mod.__dict__.items()
      if filter_mod(mod, k, v)
    })



'''
Constants that can be cached in thread-local storage
'''

from appyter.ext.functools import memcached

@memcached
def get_cwd_fs():
  ''' Return fsspec compatible chroot to the current directory
  '''
  from flask import current_app
  from appyter.ext.fsspec.core import url_to_chroot_fs
  return url_to_chroot_fs(current_app.config['CWD'])

@memcached
def get_static_fs():
  ''' Return fsspec compatible chroot to the static directory
  '''
  from flask import current_app
  from appyter.ext.fsspec.core import url_to_chroot_fs
  return url_to_chroot_fs(current_app.config['STATIC_DIR'])

@memcached
def get_input_fs():
  ''' Return fsspec compatible chroot to the static directory
  '''
  from appyter.ext.fsspec.core import url_to_chroot_fs
  return url_to_chroot_fs('storage://input/')

@memcached
def get_output_fs():
  ''' Return fsspec compatible chroot to the static directory
  '''
  from appyter.ext.fsspec.core import url_to_chroot_fs
  return url_to_chroot_fs('storage://output/')

@memcached
def _get_ipynb_io():
  ''' Return byte stream for original ipynb
  '''
  import io, shutil
  from flask import current_app
  ipynb_io = io.BytesIO()
  with get_cwd_fs().open(current_app.config['IPYNB'], 'rb') as fr:
    shutil.copyfileobj(fr, ipynb_io)
  return ipynb_io

def get_ipynb_io():
  ''' Reuse the BytesIO stream
  '''
  ipynb_io = _get_ipynb_io()
  ipynb_io.seek(0)
  return ipynb_io

@memcached
def get_nbtemplate():
  ''' Parse the ipynb
  '''
  from appyter.parse.nb import nb_from_ipynb_io
  return nb_from_ipynb_io(get_ipynb_io())

@memcached
def get_j2_env():
  ''' Get an initialized bare jinja2 environment (no context)
  '''
  from flask import current_app
  from appyter.context import get_jinja2_env
  return get_jinja2_env(config=current_app.config)

@memcached
def get_form():
  ''' Render a form for the nbtemplate
  '''
  from appyter.render.form import render_form_from_nbtemplate
  return render_form_from_nbtemplate(get_j2_env(), get_nbtemplate())

@memcached
def get_nbtemplate_json():
  ''' Render the nbtemplate as nbtemplate_json
  '''
  from appyter.render.nbinspect import render_nbtemplate_json_from_nbtemplate
  return render_nbtemplate_json_from_nbtemplate(get_j2_env(), get_nbtemplate())

@memcached
def get_fields():
  ''' Parse the nbtemplate fields
  '''
  from appyter.parse.nbtemplate import parse_fields_from_nbtemplate
  return parse_fields_from_nbtemplate(get_j2_env(), get_nbtemplate())

@memcached
def get_ipynb_hash():
  ''' A hash for the ipynb
  '''
  from appyter.ext.hashlib import sha1sum_io
  return sha1sum_io(get_ipynb_io())

@memcached
def get_html_exporer():
  ''' nbconvert html export
  '''
  from nbconvert import HTMLExporter
  html_exporter = HTMLExporter()
  html_exporter.template_name = 'classic'
  return html_exporter

@memcached
def get_base_files():
  ''' Include all (non-hidden) files in cwd (include requirements.txt, utils, etc..)
  '''
  from flask import current_app
  fs = get_cwd_fs()
  base_files = {}
  for f in fs.glob('*'):
    if f.startswith('.'): continue
    if f == current_app.config['IPYNB']: continue
    if not fs.isfile(f): continue
    base_files[f] = fs.open(f, 'rb').read()
  return base_files



import os
import click
import logging
from appyter import __version__
from appyter.cli import cli
from appyter.ext.click import click_option_setenv, click_argument_setenv
from appyter.ext.fsspec.fuse import fs_mount
from appyter.ext.asyncio.sync_contextmanager import sync_contextmanager

logger = logging.getLogger(__name__)

@cli.command(
  help='An alias for jupyter notebook which pre-fetches notebook data from a remote appyter and then launches serve',
  context_settings=dict(
    ignore_unknown_options=True,
  ),
)
@click_option_setenv('--data-dir', envvar='APPYTER_DATA_DIR', default=None, help='The directory to store data of executions')
@click_option_setenv('--cwd', envvar='APPYTER_CWD', default=os.getcwd(), help='The directory to treat as the current working directory for templates and execution')
@click_option_setenv('--host', envvar='APPYTER_HOST', default='0.0.0.0', help='The host to bind to')
@click_option_setenv('--port', envvar='APPYTER_PORT', type=int, default=5000, help='The port this server should run on')
@click.argument('args', nargs=-1, type=click.UNPROCESSED)
@click_argument_setenv('uri', envvar='APPYTER_URI', nargs=1, type=str)
@click.pass_context
def fetch_and_serve(ctx, data_dir, cwd, host, port, args, uri):
  import urllib.parse, urllib.request
  import fsspec
  # TODO: can this be less reliant on the `appyter-catalog` storage setup
  uri_parsed = urllib.parse.urlparse(uri)
  from appyter.ext.fsspec.alias import AliasFileSystemFactory
  storage_uri = f"{uri_parsed.scheme}://{uri_parsed.netloc}/storage/appyters/"
  fsspec.register_implementation('storage',
    AliasFileSystemFactory(
      'storage',
      storage_uri,
    )
  )
  # if data_dir doesn't exist, create it
  if data_dir is None: data_dir = 'tmpfs://'
  # mount the appyter into the data_dir
  with sync_contextmanager(fs_mount(data_dir, appyter=uri, cached=True)) as mnt:
    logging.info(f"Starting `appyter serve`...")
    # serve the bundle in jupyter notebook
    from appyter.helpers.serve import serve
    ctx.invoke(serve, cwd=cwd, data_dir=str(mnt), host=host, port=port, args=args)



import re
from appyter.fields import Field
from appyter.ext.re import re_full

class AutocompleteField(Field):
  ''' Representing a field that accepts a string with autocomplete.
  Auto complete will use `choices` or alternatively `file_path` -- a url to load the choices as an array.

  Be careful with this field, consider defining a constraint regex. Note that it is equivalent to
   a string field, and still accepts choices not being autocompleted.

  :param name: (str) A name that will be used to refer to the object as a variable and in the HTML form.
  :param label: (str) A human readable label for the field for the HTML form
  :param description: (Optional[str]) A long human readable description for the field for the HTML form
  :param constraint: A regular expression for validating the file name.
  :param hint: A hint to put in the field prior to content.
  :param file_path: (Optional[str]) A remote url to download the autocomplete list. Should resolve to a json list of terms.
  :param choices: (Union[List[str], Set[str], Dict[str, str]]) A set of choices that are available for this field or lookup table mapping from choice label to resulting value
  :param required: (Optional[bool]) Whether or not this field is required (defaults to false)
  :param default: (str) A default value as an example and for use during prototyping
  :param examples: (Optional[Dict[str, str]]) Named strings to provide as clickable examples
  :param section: (Optional[str]) The name of a SectionField for which to nest this field under, defaults to a root SectionField
  :param value: (INTERNAL Any) The raw value of the field (from the form for instance)
  :param \**kwargs: Additional keyword arguments used by other fields
  '''
  def __init__(self, constraint=r'.*', hint=None, **kwargs):
    super().__init__(
      constraint=constraint,
      hint=hint,
      **kwargs,
    )
    assert bool(self.choices) ^ bool(self.args.get('file_path')), 'AutocompleteField requires choices or file_path but not both'

  def constraint(self):
    if self.raw_value is None:
      return not self.args.get('required')
    elif self.choices:
      return self.raw_value in self.choices
    else:
      return re.match(re_full(self.args['constraint']), self.raw_value)



from appyter.fields import Field

class SectionField(Field):
  ''' Representing section header which other fields can be a part of
  ```python
  %%appyter hide_code
  {% do SectionField(name='section1', title='My Section', subtitle='is awesome') %}
  {% set x = IntField(name='int1', label='My Int', section='section1') %}
  ```

  :param name: (str) A name that will be used to refer to the object as a variable and in the HTML form.
  :param title: (str) A human readable title for the section
  :param subtitle: (str) A human readable subtitle for the section
  :param img: (str) The path to an image used in the field relative to the static folder
  :param \**kwargs: Additional keyword arguments used by other fields
  '''
  def __init__(self, **kwargs):
    super().__init__(**kwargs)
    self['type'] = 'section'

  @property
  def raw_value(self):
    raise Exception('DescriptionField has no value')



import json
import nbformat as nbf

def nb_from_ipynb_string(string):
  return nbf.reads(string, as_version=4)

def nb_from_ipynb_io(io):
  return nbf.read(io, as_version=4)

def nb_from_json(j):
  return nb_from_ipynb_string(json.dumps(j))

def nb_to_ipynb_string(nb):
  return nbf.writes(nb)

def nb_to_ipynb_io(nb, io):
  # remove `id` when writing, this is added by nbclient
  #  but causes nbformat validation errors when reading.
  for cell in nb.cells:
    if 'id' in cell:
      del cell['id']
  return nbf.write(nb, io)

def nb_to_json(nb):
  return json.loads(nb_to_ipynb_string(nb))



''' Each submodule represents a machanism for rendering an appyter as something new.
'''
import os
from appyter.ext.importlib import importdir
importdir(os.path.join(os.path.dirname(__file__)), __package__, globals())



import os
import json
import click
import fsspec

from appyter.cli import cli
from appyter.ext.urllib import join_url
from appyter.parse.nb import nb_from_ipynb_io
from appyter.parse.nbtemplate import parse_fields_from_nbtemplate
from appyter.context import get_env, get_jinja2_env
from appyter.ext.click import click_option_setenv, click_argument_setenv

def render_nbtemplate_json_from_nbtemplate(env, nb):
  ''' Render a json representing the relevant Fields throughout the notebook.
  '''
  return [
    { 'field': field.field, 'args': field.args }
    for field in parse_fields_from_nbtemplate(env, nb, deep=True)
  ]

@cli.command(help='Inspect appyter for arguments (fields)')
@click.option('-o', '--output', default='-', type=click.File('w'), help='The output location of the inspection json')
@click_option_setenv('--cwd', envvar='APPYTER_CWD', default=os.getcwd(), help='The directory to treat as the current working directory for templates and execution')
@click_argument_setenv('ipynb', envvar='APPYTER_IPYNB')
def nbinspect(cwd, ipynb, output, **kwargs):
  env = get_jinja2_env(config=get_env(cwd=cwd, ipynb=ipynb, mode='inspect',  **kwargs))
  with fsspec.open(join_url(cwd, ipynb), 'r') as fr:
    nbtemplate = nb_from_ipynb_io(fr)
  fields = render_nbtemplate_json_from_nbtemplate(env, nbtemplate)
  json.dump(fields, output)



import re
from appyter.fields import Field
from appyter.ext.json import try_json_loads
from appyter.ext.itertools import ensure_list

class VariableField(Field):
  ''' Represing a variable number of instances of a given Field
  Usage:
  ```j2
  {{ VariableField(
    nane='files'
    label='Files',
    field=FileField(
      name='file',
      label='File',
      default='file.txt',
      examples={},
    ),
    default=['file.txt'],
    min=0,
    max=10,
  ) }}
  ```
  :param name: (str) A name that will be used to refer to the object as a variable and in the HTML form.
  :param label: (str) A human readable label for the field for the HTML form
  :param description: (Optional[str]) A long human readable description for the field for the HTML form
  :param default: (list) A default value as an example and for use during prototyping
  :param min: (int) The minimum number of fields that can be specified
  :param max: (int) The maximum number of fields that can be specified
  :param section: (Optional[str]) The name of a SectionField for which to nest this field under, defaults to a root SectionField
  :param value: (INTERNAL Any) The raw value of the field (from the form for instance)
  :param \**kwargs: Remaining arguments passed down to :class:`appyter.fields.Field`'s constructor.
  '''
  def __init__(self, field=None, **kwargs):
    super().__init__(
      field=field,
      **kwargs,
    )
    field['args']['parent'] = self.args['name']
  #
  def prepare(self, req):
    ''' Typically, you'll just provide a list of values satisfying the Field type,
    for forms we support json pointer refs for compatibility with older form fields -- these
    pointers are resolved during prepare. Basically:

    ```json
    {
      "variablefield": [{"$ref":"#/singlefield0"},{"$ref":"#/singlefield1"}],
      "singlefield0": "a",
      "singlefield1": "b"
    }
    ```json
    Will resolve to
    ```json
    {"variablefield": ["a", "b"]}
    ```

    Letting us construct the form as before but permitting us to reference form fields
     generated on the fly that end up in the form request.
    '''
    data = super().prepare(req)
    expr = re.compile(r'^#/([^/]+)$')
    value = []
    for i, v in enumerate(ensure_list(try_json_loads(data[self.args['name']]))):
      # resolve json pointers
      if type(v) == dict and len(v) == 1 and '$ref' in v:
        ptr = v['$ref']
        m = expr.match(ptr)
        if m:
          name = m.group(1)
          # prepare from pointer to value in request
          data.update(
            self._env.globals[self.args['field']['field']](
              name=name,
            ).prepare(req)
          )
          value.append(data[name])
          continue
      # prepare with value provided using field
      name = f"{self.args['field']['args']['name']}{i}"
      data.update(
        self._env.globals[self.args['field']['field']](
          name=name,
        ).prepare({name: v})
      )
      value.append(data[name])
    # return prepared aggregated value
    self.args['value'] = value
    return { self.args['name']: value }
  #
  @property
  def raw_value(self):
    value = try_json_loads(self.args['value'])
    if value is None:
      return []
    elif type(value) == list:
      return value
    else:
      return [self.args['value']]
  #
  def constraint(self):
    ''' Number of values should satisfy bounded constraints and
    all values should satisfy the field constraints.
    '''
    return self.args['min'] <= len(self.raw_value) <= self.args['max'] and all(
      self._env.globals[self.args['field']['field']](
        name=self.args['field']['args']['name'],
        value=v
      ).constraint()
      for v in self.raw_value
    )
  #
  @property
  def value(self):
    return [
      self._env.globals[self.args['field']['field']](
        name=self.args['field']['args']['name'],
        value=v
      ).value
      for v in self.raw_value
    ]



import logging
logger = logging.getLogger(__name__)

def url_to_chroot_fs(url, pathmap=None, cached=False, appyter=None, **kwargs):
  ''' Like url_to_fs but supporting our extensions, namely:
  chroot   filesystem path is treated as the root
  pathmap  overlay other fsspec-compatible paths
  appyter  create a pathmap from an appyter ipynb
  cached   cache read/writes
  '''
  from appyter.ext.fsspec.util import split_protocol_opts
  from appyter.ext.fsspec.chroot import ChrootFileSystem
  protocol, path, opts = split_protocol_opts(url)
  # add protocol options to inner protocol
  if protocol not in kwargs: kwargs[protocol] = {}
  kwargs[protocol].update(opts)
  # ensure auto_mkdir is enabled
  if protocol == 'file':
    if 'auto_mkdir' not in kwargs[protocol]: kwargs[protocol]['auto_mkdir'] = True
  # add chroot
  fs = ChrootFileSystem(
    target_protocol=protocol,
    target_options=kwargs[protocol],
    fo=path,
  )
  # apply pathmap as needed
  if pathmap:
    from appyter.ext.fsspec.mapperfs import MapperFileSystem
    from appyter.ext.fsspec.overlayfs import OverlayFileSystem
    fs = ChrootFileSystem(
      fs=OverlayFileSystem(
        lower_fs=MapperFileSystem(pathmap=pathmap),
        upper_fs=fs,
      )
    )
  elif appyter:
    import urllib.request
    from appyter import __version__
    from appyter.ext.urllib import join_slash
    from appyter.parse.nb import nb_from_ipynb_io
    from appyter.ext.fsspec.mapperfs import MapperFileSystem
    from appyter.ext.fsspec.overlayfs import OverlayFileSystem
    # load notebook
    req = urllib.request.Request(appyter, headers={'Accept': 'application/vnd.jupyter'})
    with urllib.request.urlopen(req) as fr:
      nb = nb_from_ipynb_io(fr)
    # grab relevant files from metadata
    metadata = nb.get('metadata', {}).get('appyter', {})
    nbconstruct = metadata.get('nbconstruct', {})
    nbconstruct_version = nbconstruct.get('version', 'unknown')
    if nbconstruct_version and nbconstruct_version == __version__:
      pass
    else:
      logger.warning(f"This appyter was not created with this version, instance version was {nbconstruct_version} and our version is {__version__}. Proceed with caution")
    #
    if 'nbexecute' not in metadata:
      logger.warning('This appyter instance has not been executed, no results will be available')
    elif 'started' in metadata['nbexecute'] and 'completed' not in metadata['nbexecute']:
      logger.warning('This appyter is not finished running, no results will be available')
    elif 'started' in metadata['nbexecute'] and 'completed' in metadata['nbexecute']:
      logger.info(f"Appyter ran from {metadata['nbexecute']['started']} to {metadata['nbexecute']['completed']}")
    else:
      logger.warning('This appyter seems old, this may not work properly, please contact us and we can update it')
    #
    filename = nbconstruct.get('filename', 'appyter.ipynb')
    files = { filename: filename }
    files.update(nbconstruct.get('files', {}))
    files.update(metadata.get('nbexecute', {}).get('files', {}))
    pathmap = {
      filename: path if '://' in path else join_slash(appyter, path)
      for filename, path in files.items()
    }
    # prepare pathmaped filesystem overlayed on target_fs
    fs = ChrootFileSystem(
      fs=OverlayFileSystem(
        lower_fs=MapperFileSystem(pathmap=pathmap),
        upper_fs=fs,
      ),
    )
  if cached:
    from appyter.ext.fsspec.writecache import WriteCacheFileSystem
    fs = WriteCacheFileSystem(fs=fs)
  return fs




def dict_filter_none(d):
  return { k: v for k, v in d.items() if v }

def dict_collision_free_update(d, **kwargs):
  for k, v in kwargs.items():
    assert k not in d, 'Collision in update'
    d[k] = v



import pathlib
import tempfile
import shutil
import contextlib

@contextlib.contextmanager
def tempdir(path=None):
  ''' A context for a temporary directory that may exist,
  if it doesn't we'll use mkdtemp. Also has a safety mechanism
  when dealing with mounts -- it will not rmtree if the directory
  is a mounted directory (since this might remove files unintentially).
  '''
  tempdir = pathlib.Path(tempfile.mkdtemp(prefix='appyter-tmp') if path is None else path)
  tempdir.mkdir(parents=True, exist_ok=True)
  try:
    yield tempdir
  finally:
    if path is None and not tempdir.is_mount() and tempdir.exists():
      shutil.rmtree(tempdir)

def mktemp(suffix='', prefix='appyter-tmp', dir=None):
  ''' Same as tempfile.mktemp but with appyter prefix
  '''
  return tempfile.mktemp(suffix=suffix, prefix=prefix, dir=dir)



import multiprocessing as mp
mp.set_start_method('spawn', True)

import tempfile
import contextlib
from pathlib import Path

import appyter.ext.fsspec
from appyter.ext.asyncio.sync_contextmanager import sync_contextmanager
from appyter.ext.fsspec.fuse import fs_mount

def assert_eq(a, b): assert a == b, f"{repr(a)} != {repr(b)}"

@contextlib.contextmanager
def _test_ctx():
  with tempfile.TemporaryDirectory() as tmpdir:
    tmpdir = Path(tmpdir)
    (tmpdir/'a'/'b').mkdir(parents=True)
    with (tmpdir/'a'/'b'/'c').open('w') as fw:
      fw.write('C')
    with (tmpdir/'a'/'d').open('w') as fw:
      fw.write('D')
    with (tmpdir/'e').open('w') as fw:
      fw.write('E')
    yield tmpdir

def test_file_chroot_fuse():
  with _test_ctx() as tmpdir:
    with sync_contextmanager(fs_mount(str(tmpdir))) as fs:
      assert_eq(frozenset(str(p.relative_to(fs)) for p in fs.rglob('*')), frozenset(['a', 'a/b', 'a/b/c', 'a/d', 'e']))
      assert_eq((fs/'a'/'b'/'c').open('rb').read(), b'C')
      assert_eq((fs/'a'/'d').open('rb').read(), b'D')
      assert_eq((fs/'e').open('rb').read(), b'E')



import contextlib
import logging
from multiprocessing import Process

logger = logging.getLogger(__name__)

def _fuse_run(url, mount_point, kwargs, alias_dump):
  import fsspec.fuse
  from appyter.ext.fsspec.core import url_to_chroot_fs
  from appyter.ext.fsspec.alias import register_aliases
  register_aliases(alias_dump)
  logger.debug(f'preparing fs from {url} ({kwargs})..')
  with url_to_chroot_fs(url, **kwargs) as fs:
    logger.debug('launching fuse..')
    fsspec.fuse.run(fs, '', mount_point)
    logger.debug('teardown..')

@contextlib.asynccontextmanager
async def fs_mount(url, **kwargs):
  import os
  import signal
  import asyncio
  import pathlib
  import traceback
  from appyter.ext.fsspec.alias import dump_aliases
  from appyter.ext.asyncio.try_n_times import try_n_times
  from appyter.ext.asyncio.run_in_executor import run_in_executor
  from appyter.ext.tempfile import tempdir
  @run_in_executor
  def _assert_mounted(path):
    assert path.is_mount()
  @run_in_executor
  def _assert_not_mounted(path):
    assert not path.is_mount()
  with tempdir() as tmp:
    logger.debug(f'mounting {url} onto {tmp}')
    proc = Process(
      target=_fuse_run,
      args=(url, str(tmp), kwargs, dump_aliases()),
    )
    proc.start()
    try:
      await try_n_times(3, _assert_mounted, tmp)
      logger.debug(f"fs mount ready on {tmp}")
      yield tmp
    except Exception as e:
      logger.error(traceback.format_exc())
      raise
    finally:
      if proc.pid:
        logger.debug(f"unmounting fs from {tmp}")
        os.kill(proc.pid, signal.SIGINT) # SIGINT cleanly stops fsspec.fuse.run
        logger.debug(f"waiting for process to end")
        await asyncio.get_running_loop().run_in_executor(None, proc.join)
        await try_n_times(3, _assert_not_mounted, tmp)
    logger.debug(f"done")



import multiprocessing as mp

from appyter.ext.fsspec.mapperfs import MapperFileSystem
mp.set_start_method('spawn', True)

import tempfile
import contextlib
from pathlib import Path
from appyter.ext.asyncio.sync_contextmanager import sync_contextmanager

import appyter.ext.fsspec
from appyter.ext.fsspec.core import url_to_chroot_fs
from appyter.ext.fsspec.fuse import fs_mount

def assert_eq(a, b): assert a == b, f"{repr(a)} != {repr(b)}"

@contextlib.contextmanager
def _test_ctx():
  with tempfile.TemporaryDirectory() as tmpdir:
    tmpdir = Path(tmpdir)
    (tmpdir/'a'/'b').mkdir(parents=True)
    with (tmpdir/'a'/'b'/'c').open('w') as fw:
      fw.write('C')
    with (tmpdir/'a'/'d').open('w') as fw:
      fw.write('D')
    with (tmpdir/'e').open('w') as fw:
      fw.write('E')
    yield tmpdir

def test_file_mapperfs():
  with _test_ctx() as tmpdir:
    with MapperFileSystem(pathmap={'a/b/c': str(tmpdir/'a'/'b'/'c'), 'E': str(tmpdir/'e')}) as fs:
      assert_eq(frozenset(fs.glob('**')), frozenset(['a', 'a/b', 'a/b/c', 'E']))
      assert_eq(fs.cat('a/b/c'), b'C')
      assert_eq(fs.cat('E'), b'E')



def run_in_executor(f):
  import asyncio
  import functools
  @functools.wraps(f)
  def inner(*args, **kwargs):
    loop = asyncio.get_running_loop()
    return loop.run_in_executor(None, lambda: f(*args, **kwargs))
  return inner



import click
import logging
from appyter import __version__
from click_default_group import DefaultGroup

@click.group(cls=DefaultGroup, default='flask-app')
@click.version_option(
  prog_name='appyter',
  version=__version__,
)
def cli():
  logging.basicConfig(level=logging.INFO)



''' Endpoints for exporting results
'''
import io
import shutil
import zipfile
from flask import request, current_app, send_file, abort

from appyter.ext.fsspec.core import url_to_chroot_fs
from appyter.ext.urllib import join_url
from appyter.render.flask_app.constants import get_base_files, get_html_exporer
from appyter.render.flask_app.core import core
from appyter.ext.flask import route_join_with_or_without_slash
from appyter.parse.nb import nb_from_ipynb_io

@route_join_with_or_without_slash(core, 'export', '<path:path>', methods=['GET'])
def export(path):
  if path.endswith('/'):
    format = request.args.get('format', 'html')
    data_fs = url_to_chroot_fs(join_url('storage://output/', path))
    nbpath = current_app.config['IPYNB']
    if data_fs.exists(nbpath):
      with data_fs.open(nbpath, 'rb') as fr:
        nb = nb_from_ipynb_io(fr)
      if format == 'html':
        exporter = get_html_exporer()
        body, _rcs = exporter.from_notebook_node(nb)
        return send_file(io.BytesIO(body.encode()), mimetype='text/html', as_attachment=True, attachment_filename='output.html')
      elif format == 'zip':
        metadata = nb.get('metadata', {}).get('appyter', {})
        files = metadata.get('nbexecute', {}).get('files', metadata.get('nbconstruct', {}).get('files', {}))
        with url_to_chroot_fs('tmpfs://') as tmp_fs:
          with tmp_fs.open('output.zip', 'wb') as fw:
            with zipfile.ZipFile(fw, 'a', zipfile.ZIP_DEFLATED, False) as zf:
              for f, b in get_base_files().items():
                zf.writestr(f, b)
              for f in (nbpath, *files):
                if data_fs.exists(f):
                  with data_fs.open(f, 'rb') as fr:
                    with zf.open(f, 'w') as zfw:
                      shutil.copyfileobj(fr, zfw)
          return send_file(tmp_fs.open('output.zip', 'rb'), mimetype='application/zip', as_attachment=True, attachment_filename='output.zip')
  abort(404)



import fsspec

from appyter.ext.fsspec.writecache import WriteCacheFileSystem
fsspec.register_implementation('writecache', WriteCacheFileSystem)

from appyter.ext.fsspec.chroot import ChrootFileSystem
fsspec.register_implementation('chroot', ChrootFileSystem)

from appyter.ext.fsspec.alias import AliasFileSystemFactory
fsspec.register_implementation('tmpfs', AliasFileSystemFactory('tmpfs', 'memory://'))



import re
from appyter.fields import Field, PartialField

cell_match = re.compile(r'^(# *)?%%appyter(?P<type>.*?\n)(?P<source>.+)$', re.MULTILINE | re.DOTALL)

def parse_fields_from_nbtemplate(env, nb, deep=False):
  # catch instanciations of a field into the fields array
  fields = []
  def field_wapper(field):
    def wrapper(**kwargs):
      _field = field(**kwargs)
      fields.append(_field)
      return _field
    return wrapper
  ctx = {
    key: field_wapper(field)
    for key, field in env.globals.items()
    if isinstance(field, Field) or isinstance(field, PartialField)
  }
  # collect jinja2 sources from throughout the entire notebook
  source = '\n'.join(
    cell_m.group('source')
    for cell in nb.cells
    for cell_m in [cell_match.match(cell.source)]
    if cell_m
  )
  # render the source for its side-effect of instantiating
  #  relevant fields, thus populating the fields array
  env.from_string(source).render(ctx)
  if not deep:
    # by convention, fields which accept fields as arguments
    #  will register a `parent` field with their name allowing
    #  us to filter out these sub-fields when necessary
    fields = [
      field
      for field in fields
      if 'parent' not in field['args']
    ]
  return fields



import os
import click
from appyter.cli import cli
from appyter.context import get_env, get_jinja2_env
from appyter.ext.click import click_option_setenv, click_argument_setenv

@cli.command(help='Dockerize an appyter for maximum reproducibility')
@click.option('-o', '--output', envvar='APPYTER_OUTPUT', default='-', type=click.File('w'), help='The output location of the serialized dockerfile')
@click_option_setenv('--cwd', envvar='APPYTER_CWD', default=os.getcwd(), help='The directory to treat as the current working directory for templates and execution')
@click_argument_setenv('ipynb', envvar='APPYTER_IPYNB')
def dockerize(ipynb, cwd, output, **kwargs):
  env = get_jinja2_env(
    config=get_env(cwd=cwd, ipynb=ipynb, mode='dockerize', **kwargs),
  )
  env.get_template('production/Dockerfile.j2').stream().dump(output)



import os
import sys
import click
import fsspec
import asyncio
import datetime
import traceback
import logging
logger = logging.getLogger(__name__)

from appyter import __version__
from appyter.cli import cli
from appyter.ext.nbclient import NotebookClientIOPubHook
from appyter.parse.nb import nb_from_ipynb_io, nb_to_ipynb_io, nb_to_json
from appyter.ext.click import click_option_setenv, click_argument_setenv
from appyter.ext.urllib import join_url
from appyter.ext.fsspec.fuse import fs_mount

def cell_is_code(cell):
    return cell.get('cell_type') == 'code'

def cell_has_error(cell):
    for output in cell['outputs']:
        if output['output_type'] == 'error':
            return True
    return False

def iopub_hook_factory(nb, emit):
  async def iopub_hook(cell, cell_index):
    await emit({ 'type': 'cell', 'data': [cell, cell_index] })
  return iopub_hook

def json_emitter_factory(output):
  async def json_emitter(obj):
    import json
    print(json.dumps(obj), file=output)
  return json_emitter

async def nbexecute_async(ipynb='', emit=json_emitter_factory(sys.stdout), cwd='', subscribe=None):
  logger.info('starting')
  assert callable(emit), 'Emit must be callable'
  with fsspec.open(join_url(cwd, ipynb), 'r') as fr:
    nb = nb_from_ipynb_io(fr)
  #
  if 'appyter' not in nb.metadata:
    logger.warn('detected legacy format, upgrading..')
    nb.metadata['appyter'] = {
      'nbconstruct': {
        'version': 'unknown',
      }
    }
  #
  if 'nbexecute' not in nb.metadata['appyter']:
    nb.metadata['appyter']['nbexecute'] = {
      'version': __version__,
    }
  #
  if 'execution_info' in nb.metadata:
    logger.warn('detected legacy format, upgrading..')
    nb.metadata['appyter']['nbexecute'].update(
      version='unknown',
      started=nb.metadata['execution_info'].get('started'),
      completed=nb.metadata['execution_info'].get('completed'),
    )
    del nb.metadata['execution_info']
  #
  if 'completed' in nb.metadata['appyter']['nbexecute']:
    await emit({ 'type': 'error', 'data': f"Execution already completed at {nb.metadata['appyter']['nbexecute']['completed']}" })
    return
  elif 'started' in nb.metadata['appyter']['nbexecute']:
    await emit({ 'type': 'error', 'data': f"Execution already started at {nb.metadata['appyter']['nbexecute']['started']}" })
    return
  #
  await emit({ 'type': 'status', 'data': 'Starting' })
  #
  try:
    files = nb.metadata['appyter']['nbconstruct'].get('files')
    logger.debug(f"{cwd=} {files=}")
    async with fs_mount(cwd, pathmap=files, cached=True) as mnt:
      # setup execution_info with start time
      nb.metadata['appyter']['nbexecute']['started'] = datetime.datetime.now().replace(tzinfo=datetime.timezone.utc).isoformat()
      with (mnt/ipynb).open('w') as fw:
        nb_to_ipynb_io(nb, fw)
      #
      logger.info('initializing')
      state = dict(progress=0, status='Starting')
      if callable(subscribe):
        await subscribe(lambda: dict(nb=nb_to_json(nb), **state))
      #
      try:
        iopub_hook = iopub_hook_factory(nb, emit)
        client = NotebookClientIOPubHook(
          nb,
          allow_errors=True,
          timeout=None,
          kernel_name='python3',
          resources={ 'metadata': {'path': str(mnt) } },
          iopub_hook=iopub_hook,
        )
        await emit({ 'type': 'nb', 'data': nb_to_json(nb) })
        async with client.async_setup_kernel(
          env=dict(
            PYTHONPATH=':'.join(sys.path),
            PATH=os.environ['PATH'],
          ),
        ):
          logger.info('executing')
          state.update(status='Executing...', progress=0)
          await emit({ 'type': 'status', 'data': state['status'] })
          await emit({ 'type': 'progress', 'data': state['progress'] })
          n_cells = len(nb.cells)
          exec_count = 1
          for index, cell in enumerate(nb.cells):
            logger.debug(f"executing cell {index}")
            cell = await client.async_execute_cell(
              cell, index,
              execution_count=exec_count,
            )
            if cell_is_code(cell):
              if cell_has_error(cell):
                raise Exception('Cell execution error on cell %d' % (exec_count))
              exec_count += 1
            if index < n_cells-1:
              state['progress'] = index + 1
              await emit({ 'type': 'progress', 'data': state['progress'] })
            else:
              state['status'] = 'Success'
              await emit({ 'type': 'status', 'data': state['status'] })
      except asyncio.CancelledError:
        logger.info('cancelled')
        raise
      except Exception as e:
        logger.info('execution error')
        await emit({ 'type': 'error', 'data': str(e) })
      # Save execution completion time
      logger.info('saving')
      nb.metadata['appyter']['nbexecute']['completed'] = datetime.datetime.now().replace(tzinfo=datetime.timezone.utc).isoformat()
      # save additional files
      # TODO: in the future we should individual files and include the original urls here
      nb.metadata['appyter']['nbexecute']['files'] = {
        path: path
        for path in (
          str(p.relative_to(mnt))
          for p in mnt.rglob('*')
          if p.is_file()
        )
        if path not in (files.keys()|{ipynb})
      }
      #
      with (mnt/ipynb).open('w') as fw:
        nb_to_ipynb_io(nb, fw)
      #
      logger.info('finalized')
  except asyncio.CancelledError:
    logger.info(f"outer cancelled")
    raise
  except Exception as e:
    logger.error(traceback.format_exc())
    await emit({ 'type': 'status', 'data': 'Error initializing, try again later' })
  finally:
    logger.info('complete')
  #

@cli.command(help='Execute a jupyter notebook on the command line asynchronously')
@click.option('-o', '--output', default='-', type=click.File('w'), help='The output location to write dynamic updates')
@click_option_setenv('--data-dir', envvar='APPYTER_DATA_DIR', default=None, help='The directory that storage:// uris correspond to')
@click_option_setenv('--cwd', envvar='APPYTER_CWD', default=os.getcwd(), help='The directory to treat as the current working directory for templates and execution')
@click_argument_setenv('ipynb', envvar='APPYTER_IPYNB')
def nbexecute(ipynb, output, cwd, data_dir=None):
  if data_dir:
    import fsspec
    from appyter.ext.fsspec.alias import AliasFileSystemFactory
    fsspec.register_implementation('storage', AliasFileSystemFactory('storage', data_dir))
  import asyncio
  loop = asyncio.get_event_loop()
  loop.run_until_complete(nbexecute_async(ipynb=ipynb, emit=json_emitter_factory(output), cwd=cwd))
  loop.close()



import shutil
import logging

logger = logging.getLogger(__name__)

from fsspec import filesystem, AbstractFileSystem
from appyter.ext.urllib import join_slash, parent_url

class OverlayFileSystem(AbstractFileSystem):
  ''' OverlayFS implemented with fsspec
  '''
  root_marker = ''
  protocol = 'overlayfs'

  def __init__(self, lower_fs=None, lower_protocol=None, lower_options=dict(), upper_fs=None, upper_protocol=None, upper_options=dict(), **kwargs):
    '''
    Parameters
    ----------
    lower: str | fs | (protocol, options) initial
    upper: str | fs | (protocol, options) writes go here
    '''
    super().__init__(**kwargs)
    if not (lower_fs is None) ^ (lower_protocol is None):
      raise ValueError(
          "Please provide one of lower filesystem instance (lower_fs) or"
          " lower_protocol, not both"
      )
    if not (upper_fs is None) ^ (upper_protocol is None):
      raise ValueError(
          "Please provide one of upper filesystem instance (upper_fs) or"
          " upper_protocol, not both"
      )
    self.lower_fs = lower_fs if lower_fs is not None else filesystem(lower_protocol, **lower_options)
    self.upper_fs = upper_fs if upper_fs is not None else filesystem(upper_protocol, **upper_options)

  def __enter__(self):
    if getattr(self.lower_fs, '__enter__', None) is not None:
      self.lower_fs.__enter__()
    if getattr(self.upper_fs, '__enter__', None) is not None:
      self.upper_fs.__enter__()
    return self
  
  def __exit__(self, type, value, traceback):
    if getattr(self.upper_fs, '__exit__', None) is not None:
      self.upper_fs.__exit__(type, value, traceback)
    if getattr(self.lower_fs, '__exit__', None) is not None:
      self.lower_fs.__exit__(type, value, traceback)

  def mkdir(self, path, **kwargs):
    return self.upper_fs.mkdir(path, **kwargs)
  
  def makedirs(self, path, exist_ok=False):
    return self.upper_fs.makedirs(path, exist_ok=exist_ok)

  def rmdir(self, path):
    return self.upper_fs.rmdir(path)

  def rm(self, path, recursive=False, maxdepth=None):
    return self.upper_fs.rm(path, recursive=recursive, maxdepth=maxdepth)

  def copy(self, path1, path2, recursive=False, on_error=None, maxdepth=None, **kwargs):
    if self.upper_fs.exists(path1) or not self.lower_fs.exists(path1):
      return self.upper_fs.copy(path1, path2, recursive=recursive, on_error=on_error, maxdepth=maxdepth, **kwargs)
    else:
      if self.lower_fs.isdir(path1) and recursive:
        for f1 in self.lower_fs.walk(path1, maxdepth=maxdepth):
          f_rel = f1.replace(path1, '')
          f2_rel = join_slash(path2, f_rel)
          with self.lower_fs.open(f1, 'rb') as fr:
            self.upper_fs.makedirs(parent_url(f2_rel), exist_ok=True)
            with self.upper_fs.open(f2_rel, 'wb') as fw:
              shutil.copyfileobj(fr, fw)
      else:
        with self.lower_fs.open(path1, 'rb') as fr:
          self.upper_fs.makedirs(parent_url(path2), exist_ok=True)
          with self.upper_fs.open(path2, 'wb') as fw:
            shutil.copyfileobj(fr, fw)

  def mv(self, path1, path2, recursive=False, maxdepth=None, **kwargs):
    if self.upper_fs.exists(path1):
      return self.upper_fs.mv(path1, path2, recursive=recursive, maxdepth=maxdepth, **kwargs)
    else:
      raise PermissionError(path1)

  def exists(self, path, **kwargs):
    return self.upper_fs.exists(path, **kwargs) or self.lower_fs.exists(path, **kwargs)

  def info(self, path, **kwargs):
    logger.debug(f"info({path})")
    if self.upper_fs.exists(path) or not self.lower_fs.exists(path):
      return self.upper_fs.info(path, **kwargs)
    else:
      return self.lower_fs.info(path, **kwargs)

  def ls(self, path, detail=False, **kwargs):
    logger.debug(f"ls({path})")
    results = {}
    lower_exists = self.lower_fs.exists(path)
    upper_exists = self.upper_fs.exists(path)
    if lower_exists:
      for info in self.lower_fs.ls(path, detail=detail, **kwargs):
        if detail:
          results[info['name']] = info
        else:
          results[info] = info
    if upper_exists or not lower_exists:
      for info in self.upper_fs.ls(path, detail=detail, **kwargs):
        if detail:
          results[info['name']] = info
        else:
          results[info] = info
    return list(results.values())

  def _open(self, path, mode="rb", block_size=None, autocommit=True, cache_options=None, **kwargs):
    if 'r' not in mode:
      if 'w' in mode:
        pass
      elif not self.upper_fs.exists(path) and self.lower_fs.exists(path):
        with self.lower_fs.open(path, 'rb') as fr:
          self.upper_fs.makedirs(parent_url(path), exist_ok=True)
          with self.upper_fs.open(path, 'wb') as fw:
            shutil.copyfileobj(fr, fw)
    elif self.upper_fs.exists(path):
      pass
    elif self.lower_fs.exists(path):
      # completely readonly and only exists in lower
      return self.lower_fs._open(
        path,
        mode=mode,
        block_size=block_size,
        autocommit=autocommit,
        cache_options=cache_options,
        **kwargs
      )
    #
    return self.upper_fs._open(
      path,
      mode=mode,
      block_size=block_size,
      autocommit=autocommit,
      cache_options=cache_options,
      **kwargs
    )



import asyncio
import urllib.parse
import logging

logger = logging.getLogger(__name__)

async def setup_evaluate_notebook(emitter, job):
  from appyter.render.nbexecute import nbexecute_async
  state = dict(get_state=None)
  # state is ready -- get a reference to it
  @emitter.on('get_state')
  async def on_get_state(get_state=None, **kwargs):
    state['get_state'] = get_state
  # someone joins, send them the notebook, state
  @emitter.on('joined')
  async def on_joined(id=None, **kwargs):
    if state.get('get_state'):
      nb_state = state['get_state']()
      await emitter.emit('msg', type='nb', data=nb_state['nb'], to=id)
      await emitter.emit('msg', type='status', data=nb_state['status'], to=id)
      await emitter.emit('msg', type='progress', data=nb_state['progress'], to=id)
  #
  async def emit(data):
    await emitter.emit('msg', **data)
  #
  async def subscribe(get_state):
    await emitter.emit('get_state', get_state=get_state)
  #
  await nbexecute_async(
    cwd=job['cwd'],
    ipynb=job['ipynb'],
    emit=emit,
    subscribe=subscribe,
  )
  await emitter.emit('stopped')

async def setup_socketio(emitter, job):
  from appyter.ext.socketio import AsyncClient
  sio = AsyncClient()
  #
  @sio.event
  async def joined(data):
    if data['id'] != sio.sid:
      await emitter.emit('joined', id=data['id'])
  #
  @emitter.on('msg')
  async def on_msg(type=None, data=None, to=None, room=None, **kwargs):
    if to is not None:
      await sio.emit(type, data, priority=1, to=to, **kwargs)
    else:
      await sio.emit(type, data, priority=1, room=job['session'], **kwargs)
  #
  url = urllib.parse.urlparse(job['url'])
  await sio.connect(f"{url.scheme}://{url.netloc}", socketio_path=url.path)
  await sio.emit('join', job['session'])
  await emitter.wait('stopped')
  await emitter.flush()
  await emitter.clear()
  await sio.disconnect()

async def execute_async(job, debug=False):
  from appyter.ext.asyncio.event_emitter import EventEmitter
  emitter = EventEmitter()
  logger.debug(job)
  if 'storage' in job:
    import fsspec
    from appyter.ext.fsspec.alias import AliasFileSystemFactory
    fsspec.register_implementation('storage', AliasFileSystemFactory('storage', job['storage']))
  await asyncio.gather(
    setup_evaluate_notebook(emitter, job),
    setup_socketio(emitter, job),
  )
  logger.debug('EXITING')

def execute(job):
  debug = job.get('debug', False)
  logging.basicConfig(
    level=logging.DEBUG if debug else logging.WARNING,
    format='%(name)s %(message)s',
  )
  loop = asyncio.get_event_loop()
  loop.run_until_complete(execute_async(job, debug=debug))



import logging
import urllib.parse
from werkzeug.security import safe_join

logger = logging.getLogger(__name__)

def join_routes(*routes):
  ''' Utility function for joining routes while striping extraneous slashes
  '''
  return '/' + '/'.join([route.strip('/') for route in routes if route.strip('/')])

def secure_filepath(filepath):
  ''' Ensures this will be a sanitized relative path
  '''
  secured_filepath = safe_join('.', filepath.lstrip('/'))
  assert secured_filepath is not None, "Filepath wasn't secure"
  secured_filepath = secured_filepath[2:]
  if not secured_filepath:
    logger.debug("Filepath became empty while securing")
    return ''
  return secured_filepath

def secure_url(url):
  parsed = urllib.parse.urlparse(url)
  assert parsed.scheme in {'https', 'http', 'ftp'}, 'Invalid scheme'
  return url

def is_remote(url):
  try:
    secure_url(url)
    return True
  except:
    return False

def route_join_with_or_without_slash(blueprint, *routes, **kwargs):
  ''' Like @app.route but doesn't care about trailing slash or not
  '''
  def wrapper(func):
    routes_stripped = join_routes(*routes)
    if routes_stripped:
      blueprint.route(routes_stripped, **kwargs)(func)
    blueprint.route(routes_stripped + '/', **kwargs)(func)
    return func
  return wrapper




def sha1sum_io(io, chunk_size=65536):
  import hashlib
  sha1 = hashlib.sha1()
  while True:
    buf = io.read(chunk_size)
    if not buf: break
    sha1.update(buf)
  return sha1.hexdigest()

def sha1sum_dict(obj):
  import json, hashlib
  sha1 = hashlib.sha1()
  sha1.update(json.dumps(obj, sort_keys=True, separators=None, ensure_ascii=True).encode())
  return sha1.hexdigest()



import contextlib
import traceback
import logging

from appyter.ext.urllib import join_slash

logger = logging.getLogger(__name__)

from pathlib import PurePosixPath
from fsspec import filesystem, AbstractFileSystem
from appyter.ext.pathlib.chroot import ChrootPurePosixPath

class ChrootFileSystem(AbstractFileSystem):
  ''' chroot: update root and disallow access beyond chroot, only works on directories.
  '''
  root_marker = ''
  protocol = 'chroot'

  def __init__(self, target_protocol=None, target_options=None, fs=None, **kwargs):
    '''
    Parameters
    ----------
    target_protocol: str (optional)
      Target filesystem protocol. Provide either this or ``fs``.
    target_options: dict or None
      Passed to the instantiation of the FS, if fs is None.
    fs: filesystem instance
      The target filesystem to run against. Provide this or ``protocol``.
    '''
    super().__init__(**kwargs)
    if not (fs is None) ^ (target_protocol is None):
      raise ValueError(
          "Please provide one of filesystem instance (fs) or"
          " remote_protocol, not both"
      )
    self.kwargs = target_options or {}
    self.target_protocol = (
      target_protocol
      if isinstance(target_protocol, str)
      else (fs.protocol if isinstance(fs.protocol, str) else fs.protocol[0])
    )
    self.fs = fs if fs is not None else filesystem(target_protocol, **self.kwargs)

  def _target_fo(self):
    if '://' in self.storage_options.get('fo', ''):
      target_fo_split = self.storage_options.get('fo','').split('/')
      # [scheme://base-path][/path-part?...]
      return '/'.join(target_fo_split[:3]), '/' + '/'.join(target_fo_split[3:])
    else:
      return '', self.storage_options.get('fo', '')

  def _resolve_path(self, path):
    target_fo_base, target_fo_path = self._target_fo()
    resolved_path = join_slash(*filter(None, (target_fo_base, str((ChrootPurePosixPath(target_fo_path) / path).realpath()))))
    if resolved_path == '.':
      resolved_path = ''
    if resolved_path.startswith('./'):
      resolved_path = resolved_path[2:]
    return resolved_path

  def _unresolve_path(self, path):
    target_fo_base, target_fo_path = self._target_fo()
    if not path.startswith(target_fo_base):
      raise FileNotFoundError(path)
    path = path[len(target_fo_base):]
    # ensure leading/trailing slash is the same
    if path.startswith('/') and not target_fo_path.startswith('/'):
      path = path.lstrip('/')
    elif not path.startswith('/') and target_fo_path.startswith('/'):
      path = '/' + path
    unresolve_path = str(PurePosixPath(path).relative_to(target_fo_path))
    return unresolve_path

  @contextlib.contextmanager
  def __masquerade_os_error(self, path=None):
    try:
      yield
    except FileNotFoundError:
      if path:
        raise FileNotFoundError(path)
      else:
        raise FileNotFoundError
    except PermissionError:
      if path:
        raise PermissionError(path)
      else:
        raise PermissionError
    except OSError as e:
      logger.error(traceback.format_exc())
      if e.errno:
        raise type(e)(e.errno, f'{e.__class__.__name__} occurred, details have been hidden for security reasons')
      else:
        raise Exception(f"{e.__class__.__name__} occurred, details have been hidden for security reasons")
    except Exception as e:
      logger.error(traceback.format_exc())
      raise Exception(f"{e.__class__.__name__} occurred, details have been hidden for security reasons")

  def __enter__(self):
    if getattr(self.fs, '__enter__', None) is not None:
      with self.__masquerade_os_error():
        self.fs.__enter__()
    return self
  
  def __exit__(self, type, value, traceback):
    if getattr(self.fs, '__exit__', None) is not None:
      with self.__masquerade_os_error():
        self.fs.__exit__(type, value, traceback)

  def mkdir(self, path, **kwargs):
    with self.__masquerade_os_error(path=path):
      return self.fs.mkdir(self._resolve_path(path), **kwargs)

  def makedirs(self, path, exist_ok=False):
    with self.__masquerade_os_error(path=path):
      return self.fs.makedirs(self._resolve_path(path), exist_ok=exist_ok)

  def rmdir(self, path):
    with self.__masquerade_os_error(path=path):
      return self.upper_fs.rmdir(self._resolve_path(path))

  def rm(self, path, recursive=False, maxdepth=None):
    with self.__masquerade_os_error(path=path):
      return self.fs.rm(self._resolve_path(path), recursive=recursive, maxdepth=maxdepth)

  def copy(self, path1, path2, recursive=False, on_error=None, **kwargs):
    with self.__masquerade_os_error():
      return self.fs.copy(self._resolve_path(path1), self._resolve_path(path2), recursive=recursive, on_error=on_error, **kwargs)

  def mv(self, path1, path2, recursive=False, maxdepth=None, **kwargs):
    with self.__masquerade_os_error():
      return self.fs.mv(self._resolve_path(path1), self._resolve_path(path2), recursive=recursive, maxdepth=maxdepth, **kwargs)

  def exists(self, path, **kwargs):
    with self.__masquerade_os_error(path=path):
      return self.fs.exists(self._resolve_path(path), **kwargs)

  def info(self, path, **kwargs):
    with self.__masquerade_os_error(path=path):
      info = self.fs.info(self._resolve_path(path), **kwargs)
      return dict(info, name=self._unresolve_path(info['name']))

  def ls(self, path, detail=True, **kwargs):
    logger.debug(f"ls({path}) => fs.ls({self._resolve_path(path)})")
    results = []
    with self.__masquerade_os_error(path=path):
      for f in self.fs.ls(self._resolve_path(path), detail=detail, **kwargs):
        if detail:
          f = dict(f, name=self._unresolve_path(f['name']))
          results.append(f)
        else:
          results.append(self._unresolve_path(f))
    return results

  def _open(self, path, mode="rb", block_size=None, cache_options=None, **kwargs):
    with self.__masquerade_os_error(path=path):
      return self.fs._open(self._resolve_path(path), mode=mode, block_size=block_size, cache_options=cache_options, **kwargs)



def url_for(directory, public=False, **kwargs):
  from appyter.context import get_env
  from appyter.ext.urllib import join_url
  config = get_env()
  url = None
  if config['DEBUG']:
    # appyter debugging static serve mode in flask
    try:
      from flask import url_for as _url_for
      modified_directory = '.'.join(('__main__', directory)) if directory == 'static' else directory
      url = _url_for(modified_directory, **kwargs)
    except:
      pass
  #
  if url is None:
    # url_for outside of flask
    filename = kwargs.get('filename', kwargs.get('path'))
    assert filename is not None
    url = join_url(config['PREFIX'], directory, filename)
  #
  if public:
    # url_for public modifier -- return the public facing url
    try:
      from flask import request
      url = join_url(request.url_root, url)
    except:
      pass
  #
  if config['MODE'] == 'magic':
    # jupyter notebook magic mode
    if url.startswith('/'):
      url = url[1:]
  #
  return url




def sanitize_sha1sum(val):
  import re
  m = re.match(r'^[0-9a-f]{40}$', val)
  return val if m else None

def generate_uuid():
  import uuid
  return str(uuid.uuid4())

def sanitize_uuid(val):
  import uuid
  try:
    return str(uuid.UUID(val))
  except:
    return None



''' In the future, these will be independent of the profiles, injecting whatever needs to be.
For now, they're branches in the profile code.
'''

from appyter.cli import cli

@cli.command(help='List the available extra feature flags')
def list_extras():
  print('hide-code    Hide code by default')
  print('toc          Add a sticky table of contents to the rendered notebook')
  print('toggle-code  Add a button for toggling code visibility')
  print('catalog-integration  Add various appyter-catalog-integrated specific features')



import json
import click
from appyter.orchestration.cli import orchestration

@orchestration.command(help='A job triggered by an orchestration dispatcher queue')
@click.argument('job', type=str)
def job(job):
  from appyter.orchestration.job.job import execute
  execute(json.loads(job))



from setuptools import setup, find_packages

setup(
  name='appyter',
  python_requires='>=3.7.0',
  version=open('appyter/VERSION', 'r').read(),
  url='https://github.com/maayanLab/appyter/',
  author='Daniel J. B. Clarke',
  author_email='u8sand@gmail.com',
  long_description=open('README.md', 'r').read(),
  long_description_content_type='text/markdown; charset=UTF-8',
  license='CC-BY-NC-SA-4.0',
  classifiers=[
    "Development Status :: 4 - Beta",
    "Operating System :: OS Independent",
    "Programming Language :: Python :: 3.7",
    "Programming Language :: Python :: 3",
    "Topic :: Scientific/Engineering",
    "Topic :: Software Development :: Libraries :: Application Frameworks",
    "Framework :: Jupyter",
  ],
  install_requires=list(map(str.strip, open('requirements.txt', 'r').readlines())),
  extras_require={
    'appyter_init': 'cookiecutter',
    'production': list(map(str.strip, open('requirements.production.txt', 'r').readlines())),
  },
  packages=find_packages(exclude=('example',)),
  include_package_data=True,
  entry_points={
    'console_scripts': ['appyter=appyter.__main__:main'],
  }
)



import logging
logger = logging.getLogger(__name__)

async def run_app(config):
  import os
  import sys
  import asyncio
  return await asyncio.create_subprocess_exec(
    sys.executable, '-u', '-m', 'appyter',
    f"--socket={config['HOST']}:{config['PORT']}",
    stdout=sys.stdout,
    stderr=sys.stderr,
    env=os.environ,
  )

async def app_runner(emitter, config):
  import asyncio
  state_lock = asyncio.Lock()
  state = dict(proc=await run_app(config))
  #
  @emitter.on('reload')
  async def reload(changes):
    logger.info(f"Reload {changes}")
    async with state_lock:
      if 'proc' in state:
        proc = state.pop('proc')
        proc.kill()
        state['proc'] = await run_app(config)
  #
  @emitter.on('quit')
  async def quit():
    logger.info('Stopping app..')
    async with state_lock:
      state.pop('proc').kill()

async def app_messager(emitter, config):
  import asyncio
  from appyter.ext.urllib import join_slash
  from appyter.ext.socketio import AsyncClient
  from appyter.ext.asyncio.try_n_times import try_n_times
  sio = AsyncClient()
  #
  @emitter.on('livereload')
  async def livereload(changes):
    logger.info(f"LiveReload {changes}")
    await sio.emit('livereload', {})
  #
  @emitter.on('quit')
  async def quit():
    logger.info('Disconnecting..')
    await sio.disconnect()
  #
  origin = f"http://{config['HOST']}:{config['PORT']}"
  path = join_slash(config['PREFIX'], "socket.io")
  logger.info(f"Connecting to appyter server at {origin}{path}...")
  await asyncio.sleep(1)
  await try_n_times(3, sio.connect, origin, socketio_path=path)
  await sio.wait()

async def file_watcher(emitter, evt, path, **kwargs):
  from watchgod import awatch
  logger.info(f"Watching {path} for {evt}...")
  async for changes in awatch(path, **kwargs):
    await emitter.emit(evt, changes=changes)

def serve(app_path, **kwargs):
  import os
  import asyncio
  import appyter
  from appyter.ext.asyncio.event_emitter import EventEmitter
  from appyter.ext.watchgod.watcher import GlobWatcher
  from appyter.context import get_env
  config = get_env(**kwargs)
  loop = asyncio.get_event_loop()
  emitter = EventEmitter()
  # run the app and reload it when necessary
  loop.create_task(app_runner(emitter, config))
  # the underlying appyter library
  loop.create_task(
    file_watcher(emitter, 'reload', appyter.__path__[0],
      watcher_cls=GlobWatcher,
      watcher_kwargs=dict(
        include_dir_glob=['*'],
        include_file_glob=['*.py'],
        exclude_dir_glob=[],
        exclude_file_glob=[],
      ),
    )
  )
  # the underlying appyter library's templates/ipynb/staticfiles/...
  loop.create_task(
    file_watcher(emitter, 'livereload', os.path.join(appyter.__path__[0], 'profiles'),
      watcher_cls=GlobWatcher,
      watcher_kwargs=dict(
        include_dir_glob=['*'],
        include_file_glob=['*'],
        exclude_dir_glob=[],
        exclude_file_glob=['*.py'],
      ),
    )
  )
  # the appyter itself's filters/blueprints
  loop.create_task(
    file_watcher(emitter, 'reload', config['CWD'],
      watcher_cls=GlobWatcher,
      watcher_kwargs=dict(
        include_dir_glob=['filters', 'blueprints'],
        include_file_glob=['*.py'],
        exclude_dir_glob=[],
        exclude_file_glob=[],
      ),
    )
  )
  # the appyter itself's templates/ipynb/staticfiles/...
  loop.create_task(
    file_watcher(emitter, 'livereload', config['CWD'],
      watcher_cls=GlobWatcher,
      watcher_kwargs=dict(
        include_dir_glob=['*'],
        include_file_glob=['*'],
        exclude_dir_glob=[config['DATA_DIR']],
        exclude_file_glob=['*.py'],
      ),
    )
  )
  loop.create_task(app_messager(emitter, config))
  loop.run_forever()
  loop.run_until_complete(emitter.emit('quit'))
  loop.run_until_complete(emitter.flush())
  loop.run_until_complete(emitter.clear())



import multiprocessing as mp
mp.set_start_method('spawn', True)

import tempfile
import contextlib
from pathlib import Path

import appyter.ext.fsspec
from appyter.ext.fsspec.core import url_to_chroot_fs

def assert_eq(a, b): assert a == b, f"{repr(a)} != {repr(b)}"

@contextlib.contextmanager
def _test_ctx():
  with tempfile.TemporaryDirectory() as tmpdir:
    tmpdir = Path(tmpdir)
    (tmpdir/'a'/'b').mkdir(parents=True)
    with (tmpdir/'a'/'b'/'c').open('w') as fw:
      fw.write('C')
    with (tmpdir/'a'/'d').open('w') as fw:
      fw.write('D')
    with (tmpdir/'e').open('w') as fw:
      fw.write('E')
    yield tmpdir

def test_file_chroot_in_chroot():
  with _test_ctx() as tmpdir:
    # instantiate filesystem & check that it works
    with url_to_chroot_fs(f"chroot::file://{str(tmpdir)}") as fs:
      assert_eq(frozenset(fs.glob('**')), frozenset(['a', 'a/b', 'a/b/c', 'a/d', 'e']))
      assert_eq(fs.cat('a/b/c'), b'C')
      assert_eq(fs.cat('a/d'), b'D')
      assert_eq(fs.cat('e'), b'E')

def test_file_chroot():
  with _test_ctx() as tmpdir:
    # instantiate filesystem & check that it works
    with url_to_chroot_fs(str(tmpdir)) as fs:
      assert_eq(frozenset(fs.glob('**')), frozenset(['a', 'a/b', 'a/b/c', 'a/d', 'e']))
      assert_eq(fs.cat('a/b/c'), b'C')
      assert_eq(fs.cat('a/d'), b'D')
      assert_eq(fs.cat('e'), b'E')

def _http_serve(directory, port=8888):
  ''' Serve a simple static listing of the given directory on the specified port
  '''
  from aiohttp import web
  app = web.Application()
  app.add_routes([web.static('/subdir', str(directory), show_index=True)])
  return web.run_app(app, port=port)

async def _http_connect(url):
  ''' Successfully connect to and resolve a url
  '''
  import aiohttp
  async with aiohttp.ClientSession() as session:
    async with session.get(url) as response:
      return await response.text()

@contextlib.contextmanager
def _http_serve_ctx(directory, port=8888):
  ''' Context for _http_serve, ensuring server is ready before proceeding
  '''
  import os
  import signal
  from multiprocessing import Process
  from appyter.ext.asyncio.try_n_times import try_n_times
  from appyter.ext.asyncio.sync_coro import sync_coro
  proc = Process(
    target=_http_serve, args=(directory, port)
  )
  proc.start()
  try:
    sync_coro(try_n_times(3, _http_connect, f"http://localhost:{port}"))
    yield
  finally:
    if proc.pid:
      os.kill(proc.pid, signal.SIGINT)

def test_http_chroot():
  ''' Serve the test directory over http and ensure chroot_fs still works using the webserver
  '''
  port = 8888
  with _test_ctx() as tmpdir:
    with _http_serve_ctx(tmpdir, port):
      # instantiate filesystem & check that it works
      with url_to_chroot_fs(f"http://localhost:{port}/subdir/") as fs:
        assert_eq(fs.cat('a/b/c'), b'C')
        assert_eq(fs.cat('a/d'), b'D')
        assert_eq(fs.cat('e'), b'E')



import re
import itertools
import urllib.parse
from dataclasses import dataclass

from appyter.ext.pathlib.chroot import ChrootPurePosixPath
from appyter.ext.json import try_json_loads

def parent_url(url):
  parent, *filename = url.rsplit('/', maxsplit=1)
  return parent if filename else ''

def join_slash(*parts):
  if not parts: return ''
  part0, *parts = parts
  return '/'.join(itertools.chain((part0.rstrip('/'),), (part.lstrip('/') for part in parts)))

def join_url(root, *parts):
  return join_slash(root, str(ChrootPurePosixPath('/').join(*parts)))

def parse_qs(qs):
  ''' Convert dot-notation querystring into option params
  '''
  params = {}
  for Kv in qs.split('&'):
    K, eq, v = Kv.partition('=')
    if not eq: v = True
    _params_n_2 = None
    _params_n_1 = params
    for k in K.split('.'):
      if k not in _params_n_1: _params_n_1[k] = {}
      _params_n_2 = _params_n_1
      _params_n_1 = _params_n_2[k]
    _params_n_2[k] = try_json_loads(urllib.parse.unquote(v)) if type(v) == str else v
  return params

@dataclass
class URIParsed:
  url: str
  query: str
  fragment: str
  fragment_query: str

  @property
  def qs(self):
    if not self.query: return None
    return parse_qs(self.query)

  @property
  def fragment_qs(self):
    if not self.fragment_query: return None
    return parse_qs(self.fragment_query)

  def __str__(self):
    out = self.url
    if self.query:
      out += '?' + self.query
    if self.fragment:
      out += '#' + self.fragment
      if self.fragment_query:
        out += '?' + self.fragment_query
    elif self.fragment_query:
      out += '#?' + self.fragment_query
    return out

uri_re = re.compile(r'^(?P<url>[^\?#]+)(\?(?P<query>[^#]*))?(#(?P<fragment>[^\?]*?))?(\?(?P<fragment_query>.*?))?$')

def parse_file_uri(uri):
  return URIParsed(**uri_re.match(uri).groupdict())



import fnmatch
import logging
from watchgod.watcher import AllWatcher

logger = logging.getLogger(__name__)

class GlobWatcher(AllWatcher):
  def __init__(self, *args, **kwargs):
    self.include_file_glob = kwargs.pop('include_file_glob')
    self.include_dir_glob = kwargs.pop('include_dir_glob')
    self.exclude_file_glob = kwargs.pop('exclude_file_glob')
    self.exclude_dir_glob = kwargs.pop('exclude_dir_glob')
    super().__init__(*args, **kwargs)

  def should_watch_file(self, entry):
    included = any(
      fnmatch.fnmatch(entry.name, g)
      for g in self.include_file_glob
    )
    excluded = any(
      fnmatch.fnmatch(entry.name, g)
      for g in self.exclude_file_glob
    )
    return included and not excluded

  def should_watch_dir(self, entry):
    included = any(
      fnmatch.fnmatch(entry.path, g)
      for g in self.include_dir_glob
    )
    excluded = any(
      fnmatch.fnmatch(entry.path, g)
      for g in self.exclude_dir_glob
    )
    return included and not excluded



from pathlib import PurePosixPath

class ChrootPurePosixPath:
  ''' Similar to pathlib but guaranteed to stay within the `root` directory.
  '''
  def __init__(self, root, _path=PurePosixPath('.')):
    self.root = PurePosixPath(root)
    self.path = PurePosixPath(_path)
    assert not self.path.is_absolute()

  def relative_to(self, *other):
    return self.__class__(self.root, self.path.relative_to(*other))

  def realpath(self):
    return self.root / self.path

  def __repr__(self):
    return repr(self.path)

  def __str__(self):
    return str(self.path)

  def join(self, *others):
    path = self.path
    for other in map(PurePosixPath, others):
      parts = other.parts[1:] if other.is_absolute() else other.parts
      for p in parts:
        if p == '..': path = path.parent
        else: path = path / p
    return self.__class__(self.root, path)
    
  def __truediv__(self, other):
    return self.join(other)



import asyncio
import typing as t
from nbclient import NotebookClient
from nbformat import NotebookNode
from nbclient.util import ensure_async
from nbclient.exceptions import CellExecutionComplete, DeadKernelError, CellControlSignal

class NotebookClientIOPubHook(NotebookClient):
  ''' A notebook client with the ability to hook into iopub updates
  '''
  def __init__(self, *args, iopub_hook=None, **kwargs):
    super().__init__(*args, **kwargs)
    self.iopub_hook = iopub_hook

  async def _async_poll_output_msg(
    self,
    parent_msg_id,
    cell,
    cell_index
  ):
    assert self.kc is not None
    complete = False
    while not complete:
      msg = await ensure_async(self.kc.iopub_channel.get_msg(timeout=None))
      if msg['parent_header'].get('msg_id') == parent_msg_id:
        try:
          # Will raise CellExecutionComplete when completed
          self.process_message(msg, cell, cell_index)
        except CellExecutionComplete:
          complete = True
        finally:
          if self.iopub_hook is not None:
            await self.iopub_hook(cell, cell_index)

  def _kc_execute(self, *args, **kwargs):
    try:
      loop = asyncio.get_event_loop()
    except (RuntimeError, AssertionError):
      loop = asyncio.new_event_loop()
      asyncio.set_event_loop(loop)
    return self.kc.execute(*args, **kwargs)

  async def async_execute_cell(
      self,
      cell: NotebookNode,
      cell_index: int,
      execution_count: t.Optional[int] = None,
      store_history: bool = True) -> NotebookNode:
    """
    Executes a single code cell.

    To execute all cells see :meth:`execute`.

    Parameters
    ----------
    cell : nbformat.NotebookNode
      The cell which is currently being processed.
    cell_index : int
      The position of the cell within the notebook object.
    execution_count : int
      The execution count to be assigned to the cell (default: Use kernel response)
    store_history : bool
      Determines if history should be stored in the kernel (default: False).
      Specific to ipython kernels, which can store command histories.

    Returns
    -------
    output : dict
      The execution output payload (or None for no output).

    Raises
    ------
    CellExecutionError
      If execution failed and should raise an exception, this will be raised
      with defaults about the failure.

    Returns
    -------
    cell : NotebookNode
      The cell which was just processed.
    """
    assert self.kc is not None
    if cell.cell_type != 'code' or not cell.source.strip():
      self.log.debug("Skipping non-executing cell %s", cell_index)
      return cell

    if self.record_timing and 'execution' not in cell['metadata']:
      cell['metadata']['execution'] = {}

    self.log.debug("Executing cell:\n%s", cell.source)
    parent_msg_id = await asyncio.get_running_loop().run_in_executor(None,
      lambda: self._kc_execute(
        cell.source,
        store_history=store_history,
        stop_on_error=not self.allow_errors
      )
    )
    # We launched a code cell to execute
    self.code_cells_executed += 1
    exec_timeout = self._get_timeout(cell)

    cell.outputs = []
    self.clear_before_next_output = False

    task_poll_kernel_alive = asyncio.ensure_future(
      self._async_poll_kernel_alive()
    )
    task_poll_output_msg = asyncio.ensure_future(
      self._async_poll_output_msg(parent_msg_id, cell, cell_index)
    )
    self.task_poll_for_reply = asyncio.ensure_future(
      self._async_poll_for_reply(
        parent_msg_id, cell, exec_timeout, task_poll_output_msg, task_poll_kernel_alive
      )
    )
    try:
      exec_reply = await self.task_poll_for_reply
    except asyncio.CancelledError:
      # can only be cancelled by task_poll_kernel_alive when the kernel is dead
      task_poll_output_msg.cancel()
      raise DeadKernelError("Kernel died")
    except Exception as e:
      # Best effort to cancel request if it hasn't been resolved
      try:
        # Check if the task_poll_output is doing the raising for us
        if not isinstance(e, CellControlSignal):
          task_poll_output_msg.cancel()
      finally:
        raise

    if execution_count:
      cell['execution_count'] = execution_count
    self._check_raise_for_error(cell, exec_reply)
    self.nb['cells'][cell_index] = cell
    return cell



import logging
from appyter.render.flask_app.socketio import socketio

logger = logging.getLogger(__name__)

@socketio.on('livereload')
async def _(sid, data):
  ''' Broadcast livereload packets in development
  '''
  async with socketio.session(sid) as sess:
    if not sess['config']['DEBUG']:
      return
  logger.info('Live reloading...')
  await socketio.emit('livereload', data)



import os
import re
import sys
import click
import pathlib
import logging
from nbformat.v4 import new_code_cell
from appyter.cli import cli
from appyter.parse.nb import nb_from_ipynb_io, nb_to_ipynb_io
from appyter.ext.click import click_argument_setenv

logger = logging.getLogger(__name__)

def slugify(s):
  return re.sub(r'[^a-zA-Z0-9]+', '_', s)

def one_and_only(it):
  try:
    it = iter(it)
    ret = next(it)
    try:
      next(it)
    except StopIteration:
      pass
  except:
    raise
  return ret

@cli.command(help='Scaffold an appyter optionally with an existing Jupyter Notebook')
@click_argument_setenv('ipynb', envvar='APPYTER_IPYNB', required=False, type=click.Path(file_okay=True, path_type=str))
def init(ipynb, **kwargs):
  if ipynb is not None: ipynb = pathlib.Path(ipynb)
  try:
    from cookiecutter.main import cookiecutter
    if ipynb is None:
      cookiecutter(str(pathlib.Path(__file__).parent / 'cookiecutter'))
    else:
      if (pathlib.Path.cwd() / ipynb.stem).is_dir():
        from cookiecutter.exceptions import OutputDirExistsException
        raise OutputDirExistsException(f"Output directory '{ipynb.stem}' exists")
      cookiecutter(str(pathlib.Path(__file__).parent / 'cookiecutter'),
        extra_context={
          'project_name': ipynb.stem,
        },
      )
      nb = nb_from_ipynb_io(ipynb.open('r'))
      outdir = one_and_only(pathlib.Path.cwd().glob('*/appyter.json')).parent
      project_slug = outdir.name
      if len(nb.cells) == 0 or not nb.cells[0]['source'].startswith('#%%appyter init'):
        nb.cells.insert(0, new_code_cell(
          '#%%appyter init\nfrom appyter import magic\nmagic.init(lambda _=globals: _())'
        ))
      nb_to_ipynb_io(nb, (outdir / ipynb.name).open('w'))
  except ImportError:
    logger.error("Cookiecutter must be installed to use `appyter init`")
    sys.exit(1)
